/*
** Module   : BufferManager.java
** Abstract : Coordinates life cycle of record buffers
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050926   @23246 Created initial version. Manages multiple
**                           record buffers within each context. Maps
**                           legacy transactions to database-level
**                           transactions and registers buffers for commit
**                           and rollback notifications.
** 002 ECF 20051108   @23376 Adjust to new architecture of one Persistence
**                           instance per database. Most changes were in
**                           the inner class DBTxWrapper.
** 003 ECF 20051205   @23670 Renamed method obtain() to get() for
**                           consistency within package.
** 004 ECF 20060116   @23950 Fixed regression in initializeTransactions.
**                           Was impossible to begin a new database
**                           transaction.
** 005 ECF 20060120   @24037 Integration with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 006 ECF 20060130   @24285 Changed openScope signature. Now returns
**                           current scope depth.
** 007 ECF 20060205   @24286 Implement BatchListener and register with the
**                           TransactionManager for batchNotify events.
**                           Triggers start/end of record buffer batch
**                           assign mode.
** 008 ECF 20060428   @25869 Fixed dangling transaction defect. A database
**                           level which is not explicitly committed or
**                           rolled back by the TransactionManager is now
**                           closed by auto-rollback processing.
** 009 GES 20060517   @26197 Match new Finalizable interface requirement.
** 010 ECF 20060610   @27090 Fixed retry defect. A new database level
**                           transaction must be opened upon master
**                           transaction retries. This prevents a later
**                           commit or rollback from failing.
** 011 ECF 20060621   @27503 Added activeRecords method. Returns iterator
**                           on the records currently stored in all record
**                           buffers in the current client context for a
**                           particular database.
** 012 ECF 20060627   @27587 Modified activeRecords method. Only include
**                           non-transient records to avoid saving records
**                           which were not meant to be saved.
** 013 ECF 20060628   @27668 Added getCurrentScope method. Reports current
**                           buffer scope level.
** 014 ECF 20060719   @28110 Simplified activeRecords() implementation.
** 015 ECF 20060728   @28263 Moved DMO reference counting mechanism from
**                           RecordBuffer. incrementDMOUseCount() and
**                           decrementDMOUseCount() are now implemented
**                           here.
** 016 GES 20060804   @28431 Some javadoc improvements.
** 017 ECF 20060916   @29664 Added DMO property-level undo support. A
**                           scoped dictionary tracks all changes by DMO
**                           type, necessary to undo changes to any block
**                           scope. Also moved storage of record-level
**                           reversible actions from RecordBuffer class.
** 018 ECF 20060919   @29695 Fixed DMO eviction. The DMO implementation
**                           class must be passed to decrementDMOUseCount,
**                           because there previously was not enough
**                           information available to detect the proper
**                           DMO type when evicting objects from the
**                           persistence session.
** 019 ECF 20061010   @30307 Added support for auto-commit temp tables.
**                           Buffers which auto-commit are not managed by
**                           this class' transaction mechanism.
** 020 ECF 20061011   @30355 Rewrote logic to begin database transactions.
**                           Eliminated initializeTransactions() method
**                           and simplified scopeStart() and openScope()
**                           logic. This change closes some loopholes
**                           which would not have begun database level
**                           transactions at the correct time.
** 021 ECF 20061018   @30486 Moved processing to clean up batch assign
**                           mode from RecordBuffer. This was interfering
**                           with end-of-scope processing in some cases.
** 022 ECF 20061114   @31159 Moved batch assign mode state management from
**                           RecordBuffer to this class. Added package
**                           private methods to allow RecordBuffer access
**                           to batch mode state.
** 023 ECF 20061207   @31618 Fixed DMO validation problem. ChangeBroker
**                           needs to be marked for a pending session
**                           flush whenever a record buffer is added to
**                           the set of dirty buffers in batch assign
**                           mode.
** 024 ECF 20070130   @32037 Fixed handling of no-undo temp-tables. Change
**                           undone with a database-level transaction
**                           rollback is re-rolled forward and committed,
**                           as if the rollback had not occurred.
** 025 ECF 20070412   @32976 Fixed end of transaction lock processing.
**                           Record locks for a session are now managed by
**                           RecordLockContext.
** 026 ECF 20070414   @32993 Fixed memory leak. DBTxWrapper was holding on
**                           to record buffers after they had gone out of
**                           scope. The new locking architecture has
**                           eliminated the need for this object to hold
**                           buffer references at all.
** 027 ECF 20070502   @33368 Added lookupBuffer(). Required changes to how
**                           scoped dictionary for open buffers is managed
**                           internally.
** 028 ECF 20070503   @33390 Added code to track all defined buffers in a
**                           scoped dictionary. Added registerBuffer()
**                           method to allow new buffers to register
**                           themselves. Changed lookupBuffer() to use the
**                           new dictionary of all buffers rather than
**                           just open buffers.
** 029 ECF 20070515   @33668 Optimized startScope() slightly. Only process
**                           DBTxWrapper if beginning a full transaction.
** 030 ECF 20070601   @33916 Fixed potential memory leak. Release all
**                           reversible rollbacks at the end of a database
**                           transaction, rather than only at rollback.
** 031 ECF 20070821   @34901 Added setSilentError() method. Enables error
**                           suppression during query construction and/or
**                           initialization to mimic Progress' lenient
**                           error handling during its processing of a
**                           where clause.
** 032 ECF 20070911   @35283 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections.
** 033 ECF 20071024   @35594 Added call to allow ConnectionManager to
**                           clean up after a transaction ends. This was
**                           necessary to avoid this cleanup too early in
**                           the end-of-transaction logic flow. This
**                           problem was possible with the previous
**                           implementation, in which ConnectionManager
**                           registered itself with the TM independently.
** 034 ECF 20071029   @35632 Changed in[de]crementDMOUseCount() signature.
**                           Return use counts after change is made. Also
**                           integrated some use of generics.
** 035 ECF 20080516   @38600 Added support for dirty index update sharing.
**                           Integrated generics.
** 036 ECF 20080606   @38637 Fixed reversible rollback processing. Ensure
**                           that it only takes place when rolling back
**                           temp table transactions.
** 037 SVL 20080607   @38638 Fixed DBTxWrapper.iterate(). This method
**                           needs to perform cleanup processing for the
**                           master transaction by calling end() before
**                           begin(). [ECF: same change made for retry()].
** 038 CA  20080808   @39357 Implemented no-undo temp-table rollback - on
**                           db tx rollback, the re-roll actions are 
**                           processed for the no-undo buffers.
** 039 CA  20080826   @39568 Added rollbackPending which is called by TM
**                           when rollback targets a parent block.
** 040 ECF 20080827   @39601 Fixed lookupBuffer(). Addition of
**                           pendingBuffers logic regressed this method.
** 041 CA  20080820   @39451 Implemented DatabaseConnectionHandler - it 
**                           checks if an exception related to database
**                           connectivity errors should be re-thrown or 
**                           not. Fixed leak in allBuffers scoping.
** 042 ECF 20080925   @39945 Modified for ReversibleChain API change. That
**                           class' c'tor requires an additional argument.
** 043 ECF 20081020   @40207 Modified for Reversible API change. Method
**                           Reversible.rollback() now requires arguments.
** 044 ECF 20081118   @40548 Fixed reversible rollback processing in
**                           DBTxWrapper.rollback(). Needed a loop before
**                           actual rollback, to call Reversibles'
**                           prepareRollback() method.
** 045 ECF 20081127   @40711 Implement reclamation of temp table primary
**                           keys. On full transaction commit/rollback,
**                           keys which have been collected for recycling
**                           are made available for re-use, or discarded,
**                           respectively.
** 046 SVL 20081204   @40795 Added nestedOpen parameter to openScope()
**                           method.
** 047 ECF 20090114   @41161 Minor change to DBTxWrapper. begin() is now
**                           called from scopeStart() after all buffers
**                           have been processed. It is called from
**                           openScope() for a newly created DBTxWrapper.
** 048 ECF 20090121   @41193 Fixed defect in scopeStart()/scopeFinished().
**                           Buffers were having their enterBlock() and
**                           exitBlock() methods (respectively) invoked
**                           too often in the case of nested, open buffer
**                           scopes.
** 049 ECF 20090128   @41238 Changed activeRecords() to activeBuffers().
**                           Now returns an iterator on all active record
**                           buffers, rather than just the records they
**                           contain. The caller required additional
**                           information stored by the buffers themselves.
** 050 ECF 20090206   @41265 Added support for evicting unused DMOs in the
**                           P2J session interceptor. Exposed a view of
**                           DMOs currently in use so that the session
**                           interceptor can evict unused DMOs after a
**                           session flush. Also implemented a minor
**                           performance improvement for batch mode. We
**                           now pass the BufferManager instance to
**                           RecordBuffer.cleanupBatchMode() method to
**                           avoid an unnecessary context-local lookup.
** 051 ECF 20090218   @41330 Changed idiom for scoped maps of Reversibles.
**                           The get*() methods which return these maps
**                           will now return null if the map did not
**                           previously exist, and the caller did not set
**                           the create option to true. This avoids the
**                           unnecessary creation of maps in the common
**                           case of looking up an entry which will never
**                           be found.
** 052 ECF 20090221   @41444 Added removeAllReversibles(). Also removed
**                           unused parameter from decrementDMOUseCount().
** 053 ECF 20090315   @41608 Fixed NO-UNDO temp-table reversible rollback
**                           processing. It is necessary to prepare the
**                           reversibles for rollback before the old
**                           session is rolled back. The reversals are
**                           then executed in the new session. This allows
**                           us to initialize lazy persistent collections
**                           for extent fields before the old session is
**                           gone. This is a corner case for reversible
**                           deletes when the target DMO is also the
**                           current record in the buffer at the time of
**                           rollback.
** 054 ECF 20090610   @42632 Added evictDMOIfUnused(). Evicts DMO from
**                           Hibernate session if no resource references it.
** 055 ECF 20090623   @42962 Simplified internals and reduced memory
**                           footprint. Refactored four ScopedDictionary
**                           instances for reversible/undoable processing into
**                           one ScopedDictionary of UndoableData objects.
**                           Simplified API to access reversible/undoable
**                           action maps.
** 056 ECF 20090701   @43010 Fixed regressions introduced with #055 (@42962).
**                           getReversibles() and getUndoables() were not
**                           properly looking up or creating entries in the
**                           backing ScopedDictionary.
** 057 ECF 20090702   @43027 Fixed additional regressions from #055 (@42962).
**                           Reversible rollbacks were not being processed
**                           correctly.
** 058 ECF 20090716   @43213 Enabled record buffers to be lazily initialized.
**                           This required some refactoring with respect to
**                           scope open/close logic. Also ensure change is
**                           flushed to the database at the close of each
**                           block.
** 059 ECF 20090724   @43415 Fixed reversible/undoable mapping. Use master
**                           buffer instead of current buffer as map key. This
**                           avoids losing undo information for shared temp
**                           table buffers which go out of scope before their
**                           master temp tables. Also changed session flush
**                           strategy. We now try to flush before rollback at
**                           any level, rather than on scope finish events.
** 060 ECF 20090729   @43447 Adapted Commitable API change. Added validate()
**                           method to Commitable implementations.
** 061 CA  20090807   @43556 Reversibles must be kept by database and table
**                           name, not by RecordBuffer instance - to ensure
**                           proper reversible order when rolling back records
**                           in a table with multiple buffers defined for it.
**                           This fixes @43542.
** 062 ECF 20090810   @43582 Changed getBufferForEntity() to prevent NPE. This
**                           is needed, as RecordBuffer.getDatabaseAndTable()
**                           will now return null instead of throwing an
**                           IllegalStateException if the buffer is not yet
**                           initialized. Throw RuntimeException in various
**                           places (rather than a ConditionException), if an
**                           error is unrecoverable.
** 063 ECF 20090811   @43586 Changed decrementDMOUseCount() to prevent NPE.
**                           In a case where the thread is interrupted due to
**                           a session ending, it was possible for the
**                           increment and decrement calls to become out of
**                           balance.
** 064 ECF 20090811   @43598 Fixed regression in #061 (@43556). Reversibles
**                           were being stored by database and table name, but
**                           retrieved by RecordBuffer instance.
** 065 ECF 20090812   @43602 Changed Reversible key from String to new class
**                           BufferType. Removed getBufferForEntity(), which
**                           would fail to find a matching buffer in cases
**                           where buffers had their scopes closed before a
**                           transaction ended. A buffer lookup is no longer
**                           needed, since BufferType contains all information
**                           needed by former callers of that method.
** 066 ECF 20090816   @43665 Integrated with FieldAssigner. We get a context
**                           local Scopeable from that class, which we notify
**                           at each scope open and close event.
** 067 SVL 20091117   @44401 Added facility for releasing buffers into which a
**                           new record was loaded (or the old record was
**                           reloaded) at specific points of block processing.   
** 068 CA  20091202   @44466 Reworked the rollback mechanism, to maintain 
**                           proper reversible order and thus to solve some
**                           unique-constraint violations during rollback of
**                           some bulk delete cases and mixed record operations.
** 069 SIY 20100429   @44838 Fixed double rollback in some conditions.
** 070 CS  20121031          Removed client specific references.
** 071 OM  20130902          Added getDMOUseCount() for detecting if a dmo is refered by more 
**                           than one record buffer.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.type.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Manages all record buffers within the current user context.  This includes
 * transaction processing, buffer and block scope management.  A separate
 * instance of this class is created within each user context.  It acts as a
 * broker between the context local <code>TransactionManager</code>, the
 * persistence service layer which manages database resources, and multiple
 * record buffers opened by client code.
 * <p>
 * A record buffer registers itself with the buffer manager when client code
 * opens a new buffer scope.  If a transaction is active, that buffer is
 * immediately registered with the transaction manager for commit processing.
 * This allows the buffer to synchronize its state with the persistence
 * service when a transaction or subtransaction commit occurs.  If a
 * transaction is not active when a new buffer scope is opened, the buffer
 * manager keeps track of the open buffer, and the block scope at which it
 * was opened.  When a full transaction subsequently is opened, all buffers
 * with open scopes are batch registered for commit processing.
 * <p>
 * Also, when a full transaction begins, the buffer manager takes stock of
 * all databases referred to by open buffer scopes, and begins a new,
 * database-level transaction for each one.  The database transaction's life
 * cycle is managed by an inner class, {@link BufferManager.DBTxWrapper}.
 * This inner class registers itself for master commit and finish processing
 * with the transaction manager, such that it can reflect an application
 * level commit/rollback back to the underlying database transaction.
 * <p>
 * Updates made to existing records are tracked for undo purposes by this
 * class.  A scoped dictionary of reversible updates by DMO type is managed
 * by the context local instance of this class, for all open buffers in the
 * current client context.
 */
final class BufferManager
implements Scopeable,
           BatchListener,
           StopConditionVetoHandler,
           Resettable
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(BufferManager.class);
   
   /** Stores context local instances of this class */
   private static final ContextLocal<BufferManager> context =
   new ContextLocal<BufferManager>()
   {
      protected BufferManager initialValue() { return (new BufferManager()); }
   };
   
   /** Finalizable which clears batch mode at the appropriate times */
   private final Finalizable batchCleaner = new Finalizable()
   {
      public void finished()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
      public void iterate()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
      public void retry()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
   };
   
   /** Change broker */
   private final ChangeBroker changeBroker = ChangeBroker.get();
   
   /** Connection manager */
   private final ConnectionManager connMgr = ConnectionManager.get();
   
   /** Field assigner for function/procedure parameter assign-backs */
   private final Scopeable fieldScopes = FieldAssigner.getScopeable();
   
   /** The buffers which will be registered on the next scope entry. */
   private final Map<String, RecordBuffer> pendingBuffers = 
      new HashMap<String, RecordBuffer>();
   
   /** Scoped dictionary in which all defined record buffers are stored */
   private final ScopedDictionary<String, RecordBuffer> allBuffers =
      new ScopedDictionary<String, RecordBuffer>();

   /**
    * Scoped dictionary which key set contains all buffers into which a new
    * record was loaded (or the old record was reloaded) into corresponding
    * scopes. Value set doesn't contain any useful information and is used
    * as a temporary data storage during copying from one scope level to
    * another. 
    */
   private final ScopedDictionary<RecordBuffer, Boolean> loadedBuffers =
      new ScopedDictionary<RecordBuffer, Boolean>();

   /** Scoped dictionary in which open record buffers are stored */
   private final ScopedDictionary<String, RecordBuffer> openBuffers =
      new ScopedDictionary<String, RecordBuffer>();
   
   /** Map of records in use (in a persistence session) by database/buffer */
   private final Map<Persistence, Map<Persistable, Integer>> inUseRecords =
      new HashMap<Persistence, Map<Persistable, Integer>>();
   
   /** <code>Persistence</code> to <code>DBTxWrapper</code> instance map */
   private final Map<Persistence, DBTxWrapper> persistToDbTx =
      new IdentityHashMap<Persistence, DBTxWrapper>();
   
   /** All undoable actions scoped by block and keyed by buffer type */
   private final ScopedDictionary<BufferType, UndoData> undoData =
      new ScopedDictionary<BufferType, UndoData>();
   
   /** Set of dirty batch assign mode buffers requiring validation */
   private final Set<RecordBuffer> dirtyBatchBufs =
      new LinkedHashSet<RecordBuffer>();
   
   /** Depth at which current batch assign mode (if any) was started */
   private int batchDepth = -1;
   
   /** Flag which tracks whether silent error mode has been enabled */
   private boolean silentError = false;
   
   /**
    * Constructor; instances of this class are context-local and must be
    * accessed using the static {@link #get} method.
    */
   private BufferManager()
   {
      TransactionManager.registerBatchListener(this);
      TransactionManager.registerStopVetoHandler(this);
      TransactionManager.registerResettable(this);
      
      if (LOG.isTraceEnabled())
      {
         LOG.trace("BufferManager created in context " +
                   Utils.describeContext());
      }
   }
   
   /**
    * Retrieve the context-local instance of this class, instantiating it
    * first if necessary.
    *
    * @return  Context-local instance of this class.
    */
   static BufferManager get()
   {
      return context.get();
   }
   
   /**
    * Register with the {@link com.goldencode.p2j.util.TransactionManager
    * TransactionManager} a factory object which creates instances of this
    * class, so that they can be registered to receive notifications of
    * runtime scope start and finish events.
    * <p>
    * This method should be invoked once during the server bootstrap phase.
    */
   static void initialize()
   {
      ScopeableFactory factory = new ScopeableFactory()
      {
         public Scopeable createScopeable()
         {
            return BufferManager.get();
         }
      };
      TransactionManager.registerScopeableFactory(factory);
   }
   
   /**
    * Check the specified exception and determine whether we want to allow it
    * to be honored by the {@link TransactionManager}, or whether we want to
    * veto the STOP at this scope.
    * <p>
    * This implementation is interested only in STOP conditions which have
    * been generated by database connection errors.  STOP processing at the
    * current scope is vetoed iff the exception is an instance of {@link
    * DatabaseConnectionException} AND either of the following criteria are
    * met:
    * <ul>
    *   <li>the failing database is one that should be permanently connected
    *       to this P2J server instance;  OR
    *   <li>there is any record buffer still in scope, which is associated
    *       with the failing database.
    * </ul>
    * <p>
    * Otherwise, normal STOP processing is allowed to continue.
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if STOP should be vetoed;
    *          <code>false</code> if the STOP should be allowed. 
    */
   public boolean vetoStop(StopConditionException exc)
   {
      // If the exception was not caused by a database connection problem, it
      // it is not in our domain to veto;  allow STOP processing to continue.
      if ((exc instanceof DatabaseConnectionException) == false)
      {
         return false;
      }
      
      DatabaseConnectionException dce = (DatabaseConnectionException) exc;
      Database database = dce.getDatabase();
      
      // on database connection errors, the exception should be re-thrown if:
      
      // 1. this is a permanent database;  there is no recovery from this
      // failure
      Iterator<Database> iter = DatabaseManager.permanentConnections();
      while (iter.hasNext())
      {
         Database db = iter.next();
         if (db.equals(database))
         {
            return true;
         }
      }
      
      // 2. this is a virtual database and this block or a parent block uses
      // a table from this database
      Iterator<RecordBuffer> bufIter = allBuffers.values().iterator();
      while (bufIter.hasNext())
      {
         RecordBuffer buffer = bufIter.next();
         
         // a buffer is in use, the STOP must be vetoed
         if (buffer.isActive() && database.equals(buffer.getDatabase()))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Process a notification that a new block scope has been entered.  Notify
    * every buffer with an open scope that a new block scope has started.  If
    * the new block scope is within the context of a transaction, each buffer
    * with an open scope is registered for commit processing with the
    * transaction manager.  If the block scope represents the opening of a
    * new, full transaction, we also initialize database transactions for
    * each database referenced by a buffer with an open scope.
    * 
    * @throws  ErrorConditionException
    *          if there is an error flushing pending changes to the database
    *          as we open a new database transaction.
    */
   public void scopeStart()
   {
      allBuffers.addScope(null);
      loadedBuffers.addScope(null);
      openBuffers.addScope(null);
      undoData.addScope(null);
      fieldScopes.scopeStart();
      
      if (!pendingBuffers.isEmpty())
      {
         Iterator<RecordBuffer> iter = pendingBuffers.values().iterator();
         while (iter.hasNext())
         {
            RecordBuffer buffer = iter.next();
            registerBuffer(buffer);
            iter.remove();
         }
      }
      
      // Ensure that batch assign mode is turned off properly, regardless of
      // whether business logic invokes RecordBuffer.endBatch().  This
      // prevents a nesting error raised during RecordBuffer.startBatch(), if
      // a batch is interrupted and then a new one begins.
      TransactionManager.registerFinalizable(batchCleaner, false);
      
      int blockDepth = openBuffers.size();
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(Utils.describeContext() +
                   ":  block scope " +
                   blockDepth +
                   " started");
      }
      
      boolean inTx = TransactionManager.isTransaction();
      boolean fullTx = inTx && TransactionManager.isFullTransaction();
      
      Set<RecordBuffer> processed = new HashSet<RecordBuffer>();
      Iterator<RecordBuffer> iter = openBuffers.values().iterator();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         if (!processed.add(buffer))
         {
            // Already processed this one (buffers can be open in multiple,
            // nested scopes).
            continue;
         }
         
         try
         {
            if (inTx)
            {
               if (fullTx && buffer.isActive())
               {
                  bufferActiveInTransaction(buffer, false);
               }
               
               // Register the open buffer for commits at this block scope.
               TransactionManager.registerCommit(buffer);
            }
         }
         finally
         {
            // Notify each open buffer that a new block scope has opened.
            buffer.enterBlock(blockDepth);
         }
      }
      
      if (fullTx)
      {
         Iterator<DBTxWrapper> iter2 = persistToDbTx.values().iterator();
         while (iter2.hasNext())
         {
            iter2.next().begin();
         }
      }
   }
   
   /**
    * Process a notification that a scope is about to be exited.  Notify each
    * record buffer with an open scope that the current block scope is ending.
    */
   public void scopeFinished()
   {
      int blockDepth = openBuffers.size();
      
      // Notify all buffers that a block scope is closing.
      Set<RecordBuffer> processed = new HashSet<RecordBuffer>();
      Iterator<RecordBuffer> iter = openBuffers.values().iterator();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         if (!processed.add(buffer))
         {
            // Already processed this one (buffers can be open in multiple,
            // nested scopes).
            continue;
         }
         
         buffer.exitBlock(blockDepth);
      }
      
      undoData.deleteScope();

      // Copy the set of buffers contained in the current scope of the
      // loadedBuffers (which is about to be removed) to the enclosing scope.
      // Only buffers which are open at the enclosing or a higher scope
      // should be copied.
      if (loadedBuffers.size() > 1)
      {
         Set<RecordBuffer> currentLoadedBuffers =
                        loadedBuffers.getDictionaryAtScope(0, false).keySet();

         Collection<RecordBuffer> currentOpenBuffers =
                          openBuffers.getDictionaryAtScope(0, false).values();

         for (RecordBuffer buffer : currentLoadedBuffers)
         {
            if (!currentOpenBuffers.contains(buffer) ||
                buffer.getOpenScopeCount() > 1)
            {
               // if this buffer is missing at the current scope of open
               // buffers or it presents there but has open scope count > 1,
               // that means that it presents at a higher level and it should
               // be copied to the enclosing scope
               loadedBuffers.setValueAtScope(buffer, 1, false);
            }
         }
      }
      loadedBuffers.deleteScope();

      openBuffers.deleteScope();
      allBuffers.deleteScope();
      fieldScopes.scopeFinished();
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(Utils.describeContext() +
                   ":  block scope " +
                   blockDepth +
                   " finished");
      }
   }
   
   /**
    * Provides a notification that a batch is starting or ending.  Enters or
    * exits batch assignment mode for all record buffers in this context,
    * depending on the value of <code>start</code>.
    *
    * @param   start
    *          <code>true</code> if this is the start of a batch, otherwise
    *          <code>false</code>.
    *
    * @see     RecordBuffer#startBatch
    * @see     RecordBuffer#endBatch
    */
   public void batchNotify(boolean start)
   {
      if (start)
      {
         RecordBuffer.startBatch();
      }
      else
      {
         RecordBuffer.endBatch();
      }
   }

   /**
    * Releases the buffers into which a new record was loaded (or the old
    * record was reloaded) into the current or deeper scopes, or only clears
    * the list of buffers pending to be released.
    *
    * @param clearOnly
    *        If <code>true</code> then clear the list of buffers pending to be
    *        released, otherwise perform release of these buffers and clear
    *        the list after that.
    */
   public void resetState(boolean clearOnly)
   {
      Map<RecordBuffer, Boolean> buffers =
                                 loadedBuffers.getDictionaryAtScope(0, false);
      if (!clearOnly)
      {
         for (RecordBuffer buffer : buffers.keySet())
         {
            if (!buffer.isPendingRollbackProcessed())
               buffer.release();
            else
               buffer.resetPendingRollbackProcessed();
         }
      }

      buffers.clear();
   }

   /**
    * Register an undoable which is responsible for setting the buffer's 
    * proper current record after all reversibles are processed.
    * 
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   undoTarget
    *          The undoable responsible with setting the buffer back to its
    *          proper record.
    */
   void registerBufferUndoable(RecordBuffer buffer, Undoable undoTarget)
   {
      Map<BufferType, UndoData> dict = undoData.getDictionaryAtScope(0, true);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         ud.bufferUndoable = undoTarget.deepCopy();
      }
   }
   
   /**
    * Register a newly defined record buffer in the current scope.  Buffers
    * are deregistered implicitly when the scope in which they were
    * registered ends.
    *
    * @param   buffer
    *          Buffer to be registered.
    */
   void registerBuffer(RecordBuffer buffer)
   {
      allBuffers.addEntry(false, buffer.getDMOAlias(), buffer);
   }
   
   /**
    * Register as pending the newly defined record buffer. When the next scope
    * is entered, all the <code>pendingBuffers</code> will be registered with
    * that scope. Buffers are deregistered implicitly when the scope in which 
    * they were registered ends.
    *
    * @param   buffer
    *          Buffer pending to be registered.
    */
   void registerPending(RecordBuffer buffer)
   {
      pendingBuffers.put(buffer.getDMOAlias(), buffer);
   }
   
   /**
    * Begin batch assignment mode.  While in this mode, validation of record
    * buffers is deferred.  We remember the scope at which this occurs in
    * order to allow proper cleanup later.
    *
    * @see  RecordBuffer#startBatch
    * @see  RecordBuffer#endBatch
    * @see  RecordBuffer#cleanupBatchMode
    */
   void startBatchAssignMode()
   {
      batchDepth = getCurrentScope();
      dirtyBatchBufs.clear();
   }
   
   /**
    * End batch assignment mode if we are currently in batch assignment mode
    * and we are in the same scope at which batch assignment mode was
    * started.  Otherwise do nothing.
    *
    * @return  If the above conditions are met for batch mode to end, the
    *          set of record buffers which were modified during batch mode is
    *          returned.  Otherwise, return <code>null</code>, indicating
    *          either we are not in batch mode, or we are not in the proper
    *          scope to end it.
    *
    * @see  RecordBuffer#startBatch
    * @see  RecordBuffer#endBatch
    * @see  RecordBuffer#cleanupBatchMode
    */
   Set<RecordBuffer> endBatchAssignMode()
   {
      if (!isBatchAssignMode() || getCurrentScope() != batchDepth)
      {
         return null;
      }
      
      batchDepth = -1;
      
      return dirtyBatchBufs;
   }
   
   /**
    * Indicate whether we currently are in batch assignment mode.
    *
    * @return  <code>true</code> if in batch assignment mode, else
    *          <code>false</code>.
    *
    * @see     #startBatchAssignMode
    * @see     #endBatchAssignMode
    */
   boolean isBatchAssignMode()
   {
      return (batchDepth > -1);
   }
   
   /**
    * If we currently are in batch assignment mode, add the given record
    * buffer to the set of buffers requiring validation at the end of the
    * batch.  Otherwise, do nothing.
    *
    * @param   buffer
    *          Buffer to be added to the set of dirty batch mode buffers.
    *
    * @see     #startBatchAssignMode
    * @see     #endBatchAssignMode
    */
   void addDirtyBatchBuffer(RecordBuffer buffer)
   {
      if (isBatchAssignMode())
      {
         dirtyBatchBufs.add(buffer);
         
         // Add buffer to those requiring session flush;  necessary for DMO
         // validation if query has to go to the database.
         changeBroker.forcePendingFlush(buffer);
      }
   }
   
   /**
    * Enable or disable silent error mode.  Only enables this mode if it is
    * not already active, and only disables this mode if it is already active.
    * <p>
    * This method supports temporary error suppression during the processing
    * of query substitution parameters during construction/initialization of
    * a query.
    * 
    * @param   silentError
    *          <code>true</code> to enable silent error mode;
    *          <code>false</code> to disable.
    */
   void setSilentError(boolean silentError)
   {
      if (silentError)
      {
         // Only enable if not already enabled via this mechanism.
         if (!this.silentError && !ErrorManager.isSilentError())
         {
            ErrorManager.silentErrorEnable();
         }
      }
      else
      {
         // Only disable if previously enabled via this mechanism.
         if (this.silentError && ErrorManager.isSilentError())
         {
            ErrorManager.silentErrorDisable();
         }
      }
      
      this.silentError = silentError;
   }
   
   /**
    * Retrieve the set of reversible action objects for a given block scope.
    * <p>
    * If <code>create</code> is <code>true</code> and no set is found for the
    * given alias at the given scope, a set is created for the requested
    * scope.  Otherwise, the existing set or <code>null</code> will be
    * returned.  If newly created, a set is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no set exists in the requested scope,
    *          a new set is created and is stored for future lookups;  if
    *          <code>false</code>, no new set is created.
    *
    * @return  A set with the reversible action for the current scope, 
    *          or <code>null</code>.
    *
    * @see     Reversible
    * @see     RecordBuffer.ReversibleCreate
    * @see     RecordBuffer.ReversibleDelete
    */
   Set<Reversible> getReversibles(RecordBuffer buffer,
                                  int depth,
                                  boolean create)
   {
      Set<Reversible> set = null;
      Map<BufferType, UndoData> dict =
         undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            set = ud.reversibles;
            if (set == null && create)
            {
               set = new LinkedHashSet<Reversible>();
               ud.reversibles = set;
               ud.reversiblesMap = new HashMap<Serializable, List<Reversible>>();
            }
         }
      }
      
      return set;
   }

   
   /**
    * Retrieve a map of reversible action objects for a given block scope. The
    * reversibles are kept in a list, by DMO ID.
    * <p>
    * If <code>create</code> is <code>true</code> and no map is found for the
    * given alias at the given scope, a map is created for the requested
    * scope.  Otherwise, the existing map or <code>null</code> will be
    * returned.  If newly created, a map is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no map exists in the requested scope,
    *          a new map is created and is stored for future lookups;  if
    *          <code>false</code>, no new map is created.
    *
    * @return  A map with the reversible action for the current scope, 
    *          or <code>null</code>.
    *
    * @see     Reversible
    * @see     RecordBuffer.ReversibleCreate
    * @see     RecordBuffer.ReversibleDelete
    */
   Map<Serializable, List<Reversible>> getReversiblesMap(RecordBuffer buffer,
                                                         int depth,
                                                         boolean create)
   {
      Map<Serializable, List<Reversible>> map = null;
      Map<BufferType, UndoData> dict =
         undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            map = ud.reversiblesMap;
            if (map == null && create)
            {
               map = new HashMap<Serializable, List<Reversible>>();
               ud.reversiblesMap = map;
            }
         }
      }
      
      return map;
   }
   
   /**
    * Add a {@link Reversible} object to the reversibles set. This method also
    * ensures that a subsequent chain of reversibles will be kept, in a 
    * separate map, by DMO ID. As reversible actions for 2 DMO's can mix, this
    * map will allow access to the reversibles for a certain DMO, in the
    * proper order.
    *
    * @param   buffer
    *          Record buffer with which changes are associated.
    * @param   key
    *          The DMO ID used by this reversible.
    * @param   item
    *          Object which will perform the re-application of changes.
    */
   void addReversible(RecordBuffer buffer, Serializable key, Reversible item)
   {
      addReversible(buffer, key, item, true);
   }
   
   /**
    * Add a {@link Reversible} object to the reversibles set. When the 
    * <code>chain</chain> parameter is true, this method  ensures that a 
    * subsequent chain of reversibles will be kept, in a separate map, by DMO 
    * ID. As reversible actions for 2 DMO's can mix, this map will allow 
    * access to the reversibles for a certain DMO, in proper order.
    *
    * @param   buffer
    *          Record buffer with which changes are associated.
    * @param   key
    *          The DMO ID used by this reversible.
    * @param   item
    *          Object which will perform the re-application of changes.
    * @param   chain
    *          When true, it will add the reversible to the per-DMO map,
    *          to enable access to all reversibles registered for a DMO.
    */
   void addReversible(RecordBuffer buffer, 
                      Serializable key, 
                      Reversible item,
                      boolean chain)
   {
      if (chain)
      {
         // the key must be non-null only if we chain the reversibles in the
         // fast-access map
         assert (key != null);
      }

      Set<Reversible> set = getReversibles(buffer, 0, true);
      
      set.add(item);
      
      if (chain)
      {
         // we must add the item to the fast-access map too
         Map<BufferType, UndoData> dict = 
            undoData.getDictionaryAtScope(0, false);
         RecordBuffer master = buffer.getMasterBuffer();
         BufferType bufKey = master.getBufferType();
         
         if (dict != null)
         {
            UndoData ud = dict.get(bufKey);
            
            if (ud != null)
            {
               List<Reversible> list = ud.reversiblesMap.get(key);
               if (list == null)
               {
                  list = new ArrayList<Reversible>(3);
                  ud.reversiblesMap.put(key, list);
               }
               
               list.add(item);
            }
         }
      }
   }
   
   /**
    * Retrieve the map of DMO IDs to reversible update objects for a given
    * block scope.
    * <p>
    * If <code>create</code> is <code>true</code> and no map is found for the
    * given alias at the given scope, a map is created for the requested
    * scope.  Otherwise, the existing map or <code>null</code> will be
    * returned.  If newly created, a map is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no map exists in the requested scope,
    *          a new map is created and is stored for future lookups;  if
    *          <code>false</code>, no new map is created.
    *
    * @return  A map which associates DMO IDs with the reversible update for
    *          the current scope, or <code>null</code>.
    *
    * @see     RecordBuffer.ReversibleUpdate
    */
   Map<Serializable, Reversible> getUndoables(RecordBuffer buffer,
                                              int depth,
                                              boolean create)
   {
      Map<Serializable, Reversible> map = null;
      Map<BufferType, UndoData> dict =
         undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            map = ud.undoables;
            if (map == null && create)
            {
               map = new LinkedHashMap<Serializable, Reversible>(4);
               ud.undoables = map;
            }
         }
      }
      
      return map;
   }
   
   /**
    * Remove all <code>Reversible</code> objects associated with the given
    * buffer from their respective scoped dictionaries.
    * 
    * @param   buffer
    *          Buffer associated with the <code>Reversible</code>s to be
    *          removed.
    */
   void removeAllReversibles(RecordBuffer buffer)
   {
      undoData.deleteEntry(buffer.getMasterBuffer().getBufferType());
   }
   
   /**
    * Query the scope level which currently is open.
    *
    * @return  Current scope level.
    */
   int getCurrentScope()
   {
      return openBuffers.size();
   }
   
   /**
    * Called by {@link RecordBuffer} when a new buffer scope is opened to
    * register the open scope with the buffer manager, so that the buffer
    * can receive block scope start and stop notifications, and be registered
    * for commit processing when a full transaction is begun.
    *
    * @param   buffer
    *          Buffer for which a new scope is being opened.
    * @param   nestedOpen
    *          <code>true</code> if this buffer is already open.
    *
    * @return  Current scope depth.
    */
   int openScope(RecordBuffer buffer, boolean nestedOpen)
   {
      if (!nestedOpen)
      {
         // This buffer was not open when the block scope was started, so
         // its enterBlock method was not called;  call it now.
         int blockDepth = openBuffers.size();
         buffer.enterBlock(blockDepth);
      }
      
      openBuffers.addEntry(false, buffer.getDMOAlias(), buffer);
      
      if (TransactionManager.isTransaction())
      {
         if (buffer.isActive())
         {
            bufferActiveInTransaction(buffer, true);
         }
         
         // Register buffer for commit notifications if it has not already
         // been opened at a higher scope (in that case, the scopeStart()
         // invocation for the current scope will already have registered it).
         if (!nestedOpen)
         {
            TransactionManager.registerCommit(buffer);
         }
      }
      
      return openBuffers.size();
   }
   
   /**
    * Given a record buffer with an open scope operating within an application
    * level transaction, possibly create a new database transaction wrapper
    * object and optionally begin a database level transaction in the buffer's
    * backing database.  If a transaction wrapper already exists for the
    * buffer's database, do nothing.
    * 
    * @param   buffer
    *          Record buffer with an open buffer scope.
    * @param   beginTx
    *          Indicates whether to begin a new database transaction, in
    *          addition to creating a transaction wrapper.  This parameter is
    *          ignored if a transaction wrapper already exists.
    */
   void bufferActiveInTransaction(RecordBuffer buffer, boolean beginTx)
   {
      Persistence persistence = buffer.getPersistence();
      DBTxWrapper dbTx = persistToDbTx.get(persistence);
      
      // If dbTx is null, we have to create one which is associated with
      // this buffer's Persistence object.
      if (dbTx == null)
      {
         // If we have any pending temp-table updates not yet flushed to the
         // database, we need to make sure they are flushed now, so that they
         // cannot be rolled back if the transaction we are about to open is
         // rolled back. This can happen, for instance, if an update is made
         // to a newly created record which was flushed to the session soon
         // after creation.
         if (buffer.isTemporary())
         {
            try
            {
               persistence.maybeFlush();
            }
            catch (PersistenceException exc)
            {
               // Database problem.
               throw new RuntimeException(exc);
            }
         }
         
         dbTx = new DBTxWrapper(persistence, buffer.getDirtyContext());
         
         if (beginTx)
         {
            dbTx.begin();
         }
      }
   }
   
   /**
    * Find the record buffer associated with the given alias, which is the
    * variable name with which the buffer was defined.
    *
    * @param   alias
    *          Buffer variable name in business logic.
    *
    * @return  Associated record buffer or <code>null</code> if not found.
    */
   RecordBuffer lookupBuffer(String alias)
   {
      RecordBuffer buffer = pendingBuffers.get(alias);
      if (buffer == null)
      {
         buffer = allBuffers.lookup(alias);
      }
      
      return buffer;
   }

   /**
    * Get an iterator on all open buffers which contain non-<code>null</code>
    * current records for the given persistence object, within the current
    * client context.  Auto-commit record buffers associated with a pending,
    * reversible rollback are not included in the iteration.  Records that
    * represent copies of DMOs from other sessions, obtained from a {@link
    * DirtyShareContext}, are likewise excluded.
    *
    * @param   persistence
    *          Persistence object which manages sessions.  Only record
    *          buffers associated with this persistence object are returned.
    *
    * @return  Iterator on all active buffers.
    */
   Iterator<RecordBuffer> activeBuffers(Persistence persistence)
   {
      Set<RecordBuffer> set = new HashSet<RecordBuffer>();
      Iterator<RecordBuffer> iter = openBuffers.values().iterator();
      while (iter.hasNext())
      {
         RecordBuffer next = iter.next();
         if (next.isActive() &&
             !next.isTransient() &&
             !next.isDirtyCopy() &&
             persistence.equals(next.getPersistence()))
         {
            Persistable dmo = next.getCurrentRecord();
            if (dmo != null)
            {
               set.add(next);
            }
         }
      }
      
      return set.iterator();
   }
   
   /**
    * Get a read-only view of the DMOs currently in use, associated with the
    * given <code>Persistence</code> instance. This method may create the
    * backing DMO-to-use-count map by side effect.
    * 
    * @param   persistence
    *          Persistence object which manages sessions.
    * 
    * @return  Set of active DMOs in this user context, associated with the
    *          given <code>Persistence</code> instance.
    */
   Set<Persistable> getInUseDMOs(Persistence persistence)
   {
      Map<Persistable, Integer> map = getDMOInUseMap(persistence);
      
      return Collections.unmodifiableSet(map.keySet());
   }
   
   /**
    * Increment the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #decrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    *
    * @param   persistence
    *          Persistence object which manages sessions.
    * @param   dmo
    *          DMO instance whose reference count is to be incremented.
    * 
    * @return  Use count for <code>dmo</code> after incrementing.
    */
   int incrementDMOUseCount(Persistence persistence, Persistable dmo)
   {
      Map<Persistable, Integer> map = getDMOInUseMap(persistence);
      Integer count = map.get(dmo);
      int i = (count == null ? 1 : count.intValue() + 1);
      map.put(dmo, new Integer(i));
      
      return i;
   }

   /**
    * Return the context local use count for the specific instance of the given record.
    * Used in conjunction with {@link #decrementDMOUseCount} and {@link #incrementDMOUseCount} to
    * maintain an evidence of the buffers referring the dmo. Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * The actual count of buffers referring the dmo is used to emulate the P4GL behavior when
    * <code>rereadnolock</code> is specified. When this parameter is active, the dmo is refreshed
    * from database no matter how many buffers refer the dmo, otherwise only refresh the dmo if
    * it is referred by only one buffer.
    *
    * @param   persistence
    *          Persistence object which manages sessions.
    * @param   dmo
    *          DMO instance whose reference count is to be incremented.
    *
    * @return  Use count for <code>dmo</code> after incrementing.
    */
   int getDMOUseCount(Persistence persistence, Persistable dmo)
   {
      Map<Persistable, Integer> map = inUseRecords.get(persistence);
      if (map == null)
      {
         return 0;
      }
      Integer count = map.get(dmo);
      return count == null ? 0 : count.intValue();
   }
   
   /**
    * Decrement the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #incrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    *
    * @param   persistence
    *          Persistence object which manages sessions.
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  Use count for <code>dmo</code> after decrementing, or -1 if use
    *          count could not be found (possible in rare cases of interrupted
    *          thread on session exit).
    *
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    */
   int decrementDMOUseCount(Persistence persistence, Persistable dmo)
   throws PersistenceException
   {
      Map<Persistable, Integer> map = inUseRecords.get(persistence);
      Integer count = map.remove(dmo);
      if (count == null)
      {
         return -1;
      }
      
      int i = count - 1;
      if (i == 0)
      {
         persistence.detach(dmo);
      }
      else
      {
         map.put(dmo, i);
      }
      
      return i;
   }
   
   /**
    * Evict the given DMO from the current Hibernate session if it is not
    * referenced by any buffer or other resource.  This method does not
    * modify the DMO's use count, if any.  It is intended to quickly evict
    * DMOs which have been loaded into the session by side effect.
    * 
    * @param   persistence
    *          Persistence object which manages sessions.
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  <code>true</code> if the DMO was evicted;  <code>false</code>
    *          if it was referenced by other resources.
    *
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    */
   boolean evictDMOIfUnused(Persistence persistence, Persistable dmo)
   throws PersistenceException
   {
      Map<Persistable, Integer> map = inUseRecords.get(persistence);
      if (map == null || !map.containsKey(dmo))
      {
         persistence.detach(dmo);
         
         return true;
      }
      
      return false;
   }

   /**
    * Notify the buffer manager that a new record was loaded (or the old
    * record was reloaded) into the specified record buffer. 
    *
    * @param recordBuffer
    *        The record buffer in which a record was loaded (or the old record
    *        was reloaded). 
    */
   void notifyRecordWasLoaded(RecordBuffer recordBuffer)
   {
      if (recordBuffer.getCurrentRecord() == null ||
          recordBuffer.isTemporary())
      {
         return;
      }

      loadedBuffers.addEntry(false, recordBuffer, false);
   }

   /**
    * Get the map of active DMOs in this user context to their use counts,
    * which is associated with the given <code>Persistence</code> instance.
    * The map is created if necessary.
    * 
    * @param   persistence
    *          Persistence object which manages sessions.
    * 
    * @return  Map of active DMOs in this user context to their use counts,
    *          associated with the given <code>Persistence</code> instance.
    */
   private Map<Persistable, Integer> getDMOInUseMap(Persistence persistence)
   {
      Map<Persistable, Integer> map = inUseRecords.get(persistence);
      if (map == null)
      {
         map = new IdentityHashMap<Persistable, Integer>();
         inUseRecords.put(persistence, map);
      }
      
      return map;
   }
   
   /**
    * Compose a message suitable for logging.  Reports buffer scope and
    * context information in addition to a custom message.
    *
    * @param   persistence
    *          Persistence object which will provide context info.
    * @param   text
    *          Custom message text.
    *
    * @return  Composed message text.
    */
   private String message(Persistence persistence, String text)
   {
      StringBuilder buf = new StringBuilder("scope ");
      buf.append(openBuffers.size());
      buf.append(": ");
      buf.append(text);
      
      return persistence.message(buf.toString());
   }
   
   /**
    * Storage mechanism for maps of reversible/undoable actions, keyed by
    * primary key.  One instance of this class holds all such objects for a
    * single buffer in a single scope.
    */
   private static class UndoData
   {
      /** Reversible record actions linked in natural order */
      Set<Reversible> reversibles;

      /** List of reversible record actions mapped by ID, in natural order */
      Map<Serializable, List<Reversible>> reversiblesMap;
      
      /** Undoable record actions (or rollbacks) keyed by primary key */
      Map<Serializable, Reversible> undoables;
      
      /** 
       * Undoable which will set the no-undo buffer to its proper record 
       * after the reversibles are processed.
       */
      Undoable bufferUndoable;
      
      /**
       * Indicate whether this data structure holds any data.
       * 
       * @return  <code>true</code> if this object holds any reversible or
       *          undoable actions, else <code>false</code>.
       */
      boolean isEmpty()
      {
         return (reversibles == null && undoables == null);
      }
   }
   
   /**
    * A helper class which is created whenever a new database-level
    * transaction is needed.  Upon creation, this object opens a new
    * database-level transaction, and registers itself with the
    * <code>TransactionManager</code> to receive commit/rollback and
    * finished/iterate/retry notifications.
    * <p>
    * Note that our architecture differs somewhat from Progress here in that
    * if we have an error at commit time, it is too late to undo in the
    * current block scope.  All attempts are made before commit is executed
    * to vet and prevent any errors, but it is still possible to have an
    * error at this level if say, two transactions in different contexts
    * attempt to commit conflicting data.  In this case, the error would be
    * handled in the next block scope up from the current one.
    * <p>
    * Auto-commit buffers require special handling during a database-level
    * rollback.  Changes to such buffers made during a database transaction
    * must be preserved, but a database-level transaction rollback will not
    * discriminate these changes from those made to other buffers.  Therefor,
    * we must reverse the rollback (re-roll forward?) for any auto-commit
    * buffers whose changes are rolled back with the transaction.  This is
    * done by beginning a new transaction just for these changes, immediately
    * after a database-level rollback.  Within that transaction, changes made
    * to an auto-commit buffer are rolled forward once again and the
    * transaction is then committed.
    */
   private class DBTxWrapper
   implements Finalizable,
              Commitable
   {
      /** Persistence service object */
      private final Persistence persistence;
      
      /** Dirty share context */
      private final DirtyShareContext dirtyContext;
      
      /** Do we need to rollback database transaction in finished()? */
      private boolean autoRollback = false;
      
      /**
       * Constructor which accepts a {@link Persistence} instance and the set
       * of currently open buffers for that database, registers with the
       * <code>TransactionManager</code> as a <code>Finalizable</code> and a
       * <code>Commitable</code>, and begins a new database-level
       * transaction.
       *
       * @param   persistence
       *          Persistence object which will be used for transaction
       *          support.
       * @param   dirtyContext
       *          Dirty share context which must be cleaned up at the end of
       *          this transaction.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      DBTxWrapper(Persistence persistence, DirtyShareContext dirtyContext)
      {
         if (persistence == null)
         {
            throw new NullPointerException();
         }
         
         persistToDbTx.put(persistence, this);
         
         this.persistence = persistence;
         this.dirtyContext = dirtyContext;
         
         // The SessionFlusher must be the first (regular) Commitable
         // registered, because it must do its work before any other
         // Commitable is called for the same scope.
         SessionFlusher flusher = new SessionFlusher(persistence);
         TransactionManager.registerCommit(flusher);
         
         // Register for master transaction commit/rollback.
         TransactionManager.registerTransactionCommit(this, true);
         
         // Register for master transaction finish/iterate.
         TransactionManager.registerTransactionFinish(this, true);
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug(
               message(persistence,
                       "registered for master transaction commit/finished"));
         }
      }
      
      /**
       * Called by the <code>TransactionManager</code> after commit or
       * rollback to clean up.  This closes the current Hibernate session
       * and removes this database name from the buffer manager's set of
       * databases currently involved in transactions.
       * <p>
       * In the event the transaction manager did not explicitly commit or
       * rollback (e.g., <code>blockSetup()</code> was never called, perhaps
       * due to conditional processing which evaluated to false), we
       * automatically rollback at this point.
       */
      public void finished()
      {
         try
         {
            if (autoRollback)
            {
               if (LOG.isDebugEnabled())
               {
                  LOG.debug(message(persistence, "triggering auto-rollback"));
               }
               
               rollback(true);
            }
         }
         finally
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug(message(persistence, "master transaction finished"));
            }
            
            persistToDbTx.remove(persistence);
            end();
         }
      }
      
      /**
       * Called by the <code>TransactionManager</code> from the top of the
       * block (just after the loop test) on the second and subsequent
       * iterations of an iterating loop.
       * <p>
       * We create a new, database-level transaction here, after the loop
       * test has been passed.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      public void iterate()
      {
         if (LOG.isDebugEnabled())
         {
            LOG.debug(message(persistence, "master transaction iterate"));
         }
         
         end();
         begin();
      }
      
      /**
       * Processes the retry notification provided by the
       * <code>TransactionManager</code>.  This occurs just after we've
       * rolled back and we are about to jump to the top of the loop.  We
       * will not get a call to {@link #iterate}, because this is not a
       * normal iteration;  instead, we are retrying the same pass through
       * the block and need to set up a transaction as if we were coming in
       * fresh.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      public void retry()
      {
         if (LOG.isDebugEnabled())
         {
            LOG.debug(message(persistence, "master transaction retry"));
         }
         
         end();
         begin();
      }
      
      /**
       * Called by the <code>TransactionManager</code> when the full
       * transaction must be committed.  This is our hook to commit the
       * database-level transaction.
       *
       * @param   transaction
       *          <code>true</code> if this is a full transaction.  Not used,
       *          since we only register for full transaction commits.
       */
      public void commit(boolean transaction)
      {
         if (LOG.isDebugEnabled())
         {
            LOG.debug(message(persistence, "master transaction commit"));
         }
         
         try
         {
            autoRollback = false;
            persistence.commit();
         }
         catch (PersistenceException exc)
         {
            if (LOG.isErrorEnabled())
            {
               LOG.error(message(persistence, "commit error"), exc);
            }
            
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
         finally
         {
            // Reclaim any primary keys which had been waiting to be recycled,
            // pending a commit.
            Database db = persistence.getDatabase();
            if (DatabaseManager.TEMP_TABLE_DB.equals(db))
            {
               Commitable ctx = TemporaryBuffer.getContextCommitable();
               ctx.commit(transaction);
            }
         }
      }
      
      /**
       * No-op
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       * 
       * @param   transaction
       *          Not used.
       */
      public void validate(boolean transaction)
      {
      }
      
      /**
       * Called by the <code>TransactionManager</code> when the full
       * transaction must be rolled back.  This is our hook to roll back the
       * database-level transaction.
       *
       * @param   transaction
       *          <code>true</code> if this is a full transaction.  Not used,
       *          since we only register for full transaction commits.
       */
      public void rollback(boolean transaction)
      {
         if (LOG.isDebugEnabled())
         {
            LOG.debug(message(persistence, "master transaction rollback"));
         }
         
         Database database = persistence.getDatabase();
         boolean isTempDB = DatabaseManager.TEMP_TABLE_DB.equals(database);
         
         try
         {
            autoRollback = false;
            persistence.rollback();
         }
         catch (PersistenceException exc)
         {
            if (LOG.isErrorEnabled())
            {
               LOG.error(message(persistence, "rollback error"), exc);
            }
            
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
         
         if (isTempDB)
         {
            // do this after the session rolledback to load the reversibles 
            // in the same state as before the transaction was begun
            prepareReversibleRollbacks();
            
            // Clear any primary keys which had been waiting to be recycled,
            // pending a commit.
            Commitable ctx = TemporaryBuffer.getContextCommitable();
            ctx.rollback(transaction);
            
            // If any auto-commit buffer changes were rolled back at the
            // database, we must reverse these now.
            executeReversibleRollbacks();
         }
      }
      
      /**
       * Begin a new database-level transaction.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      private void begin()
      {
         try
         {
            if (!persistence.beginTransaction())
            {
               throw new IllegalStateException(
                  message(persistence,
                          "database transaction already active"));
            }
            
            autoRollback = true;
         }
         catch (PersistenceException exc)
         {
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
      }
      
      /**
       * Perform transaction post-processing to release or downgrade record
       * locks which may have been acquired during the transaction, to clean
       * up dirty index update sharing information contributed to the dirty
       * share manager by the current context for the transaction which is now
       * ending, and to clean up any other resources associated with the
       * transaction.
       */
      private void end()
      {
         // Clean up dirty shares for this transaction.
         if (dirtyContext != null)
         {
            dirtyContext.cleanup();
         }
         
         // Remove all pending reversible rollbacks associated with
         // this database.
         Database database = persistence.getDatabase();
         Iterator<Map.Entry<BufferType, UndoData>> iter =
            undoData.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<BufferType, UndoData> next = iter.next();
            BufferType key = next.getKey();
            
            if (database.equals(key.getDatabase()))
            {
               UndoData ud = next.getValue();
               ud.reversibles = null;
               ud.reversiblesMap = null;
               if (ud.isEmpty())
               {
                  iter.remove();
               }
            }
         }
         
         // Release or downgrade record locks which may have been acquired
         // during the transaction.
         RecordLockContext lockCtx = persistence.getRecordLockContext();
         lockCtx.transactionEnded();
         
         // Allow the connection manager to perform its end of transaction
         // cleanup.  This must occur after all transaction processing,
         // because it may result in tearing down the database's session
         // factory.
         connMgr.transactionEnded();
      }
      
      /**
       * Prepare temp table reversible rollbacks for execution.
       */
      private void prepareReversibleRollbacks()
      {
         Iterator<UndoData> iter = new ReversibleIterator();
         while (iter.hasNext())
         {
            UndoData data = iter.next();
            
            if (data.reversibles == null)
            {
               continue;
            }
            
            Iterator<Reversible> revIter = data.reversibles.iterator();
            while (revIter.hasNext())
            {
               Reversible rev = revIter.next();
               
               rev.prepareRollback(null);
            }
         }
      }
      
      /**
       * Reverse temp table changes which were rolled back at the database, in
       * order to support the Progress temp-table NO-UNDO semantic.
       */
      private void executeReversibleRollbacks()
      {
         ReversibleIterator iter = new ReversibleIterator();
         while (iter.hasNext())
         {
            UndoData data = iter.next();
            
            try
            {
               Set<Reversible> set = data.reversibles;
               
               if (set != null && persistence.beginTransaction())
               {
                  Iterator<Reversible> revIter = set.iterator();
                  while (revIter.hasNext())
                  {
                     Reversible rev = revIter.next();
                     rev.rollback(true, true);
                     revIter.remove();
                     
                     // after its done processing, remove it from the fast 
                     // access map and change the DMO for all its right-side
                     // siblings. this is needed to ensure that the subsequent
                     // reversibles will use the same DMO as this one.
                     try
                     {
                        List<Reversible> list = 
                           data.reversiblesMap.get(rev.getId());
                        
                        list.remove(rev);
                        Iterator<Reversible> listIter = list.iterator();
                        while (listIter.hasNext())
                        {
                           Reversible sibling = listIter.next();
                           Persistable dmo = rev.getDMO();
                           sibling.setDMO(dmo);
                        }
                     }
                     catch (UnsupportedOperationException exc) 
                     {
                        // ignore, is thrown only by bulk delete reversibles
                        // when getId() is called
                     }
                  }
                  
                  persistence.commit();
               }
            }
            
            catch (PersistenceException exc)
            {
               if (LOG.isErrorEnabled())
               {
                  LOG.error(
                     message(persistence, "reversible rollback error"),
                     exc);
               }
               
               // No good way to recover from this.  Let it abend.
               throw new RuntimeException(exc);
            }
            
            finally
            {
               if (data.bufferUndoable != null)
               {
                  Undoable u = data.bufferUndoable;
                  // this ensures the buffer remains set on its correct record
                  u.assign(u);
               }
            }
         }
      }
      
      /**
       * Class used to iterate through all the reversible data, by buffer.
       */
      private class ReversibleIterator
      implements Iterator<UndoData>
      {
         private Iterator<Map.Entry<BufferType, UndoData>> outerIter =
            undoData.entrySet(0).iterator();
         
         private UndoData nextUndoData = null;

         public boolean hasNext()
         {
            fetchNextData();
            
            return (nextUndoData != null);
         }
         
         public UndoData next()
         {
            if (nextUndoData == null)
            {
               throw new NoSuchElementException();
            }
            
            UndoData data = nextUndoData;
            nextUndoData = null;

            return data;
         }
         
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
         
         private void fetchNextData()
         {
            while (nextUndoData == null && outerIter.hasNext())
            {
               Map.Entry<BufferType, UndoData> next = outerIter.next();
               
               BufferType key = next.getKey();
               
               if (key.isUndoable())
               {
                  continue;
               }
               
               nextUndoData = next.getValue();
            }
         }
      }
   }
   
   /**
    * Commitable implementation which tries to flush Hibernate session before
    * any rollback.  Must be the first <code>Commitable</code> registered at a
    * transaction scope (i.e., before an record buffers).
    */
   private static class SessionFlusher
   implements Commitable
   {
      /** Persistence service object */
      private final Persistence persistence;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence service object.
       */
      SessionFlusher(Persistence persistence)
      {
         this.persistence = persistence;
      }
      
      /**
       * Called when <code>TransactionManager</code> processes a commit event.
       * <p>
       * This implementation does nothing.
       *
       * @param    transaction
       *           <code>true</code> if this is a full transaction and
       *           <code>false</code> if this is only a sub-transaction (a
       *           nested scope with transaction support).
       */
      public void commit(boolean transaction)
      {
      }
      
      /**
       * Called when <code>TransactionManager</code> processes a rollback
       * event.
       * <p>
       * Attempt to flush the Hibernate session if necessary.
       *
       * @param    transaction
       *           <code>true</code> if this is a full transaction and
       *           <code>false</code> if this is only a sub-transaction (a
       *           nested scope with transaction support).
       */
      public void rollback(boolean transaction)
      {
         try
         {
            persistence.maybeFlush();
         }
         catch (PersistenceException exc)
         {
            throw new RuntimeException(exc);
         }
      }
      
      /**
       * No-op
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       */
      public void validate(boolean transaction)
      {
      }
   }
}
