/*
** Module   : PreselectQuery.java
** Abstract : A query which enables stateful, cursored navigation of a server
**            side result set
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20051115   @23396 Created initial version, based on prototype
**                           CursoredQuery. Backing support for queries
**                           which must support the Progess, row-level
**                           access semantic against a preselected result
**                           set.
** 002 ECF 20051129   @23575 Added lenient-off-end mode. When in this
**                           mode, QueryOffEndExceptions thrown by the
**                           underlying buffer during next and previous
**                           operations are suppressed.
** 003 ECF 20051216   @23750 Added scrolling support. Modified retrieval
**                           logic to allow for repositioning. Added many
**                           new methods.
** 004 ECF 20060105   @23831 Inserted abstraction layer between query and
**                           its scrollable results. Added the Results
**                           interface and its default implemenation,
**                           SimpleResults. This allows subclasses to
**                           provide custom implementations by overriding
**                           the createResults method.
** 005 ECF 20060202   @24358 Added temp table support. Use buffer's
**                           multiplex ID as a query substitution
**                           parameter when executing HQL. Parameters are
**                           now managed separately, by component, to
**                           enable this processing.
** 006 ECF 20060208   @24721 Added support for client-side where clause
**                           processing. Results and SimpleResults are now
**                           top level classes. Modified all single-table
**                           constructor variants to accept an instance of
**                           WhereExpression. Modified Component to handle
**                           argument resolution more correctly. Results
**                           can now be provided by an external entity
**                           (CompoundQuery typically).
** 007 ECF 20060302   @25019 Add support for foreign key joins. Added new
**                           constructor and addComponent method variants
**                           which mirror existing versions, but accept an
**                           additional, inverse DMO parameter. Updated
**                           Component class, HQL generation, and query
**                           execution to integrate natural joins.
** 008 ECF 20060422   @25657 Added setIterationType method. Sets retrieval
**                           mode (first, last, next, etc.) for the most
**                           recently added query component.
** 009 ECF 20060423   @25661 Allow multiple client-side where clause
**                           expressions to be registered. This is needed
**                           because each table in a multi-table query can
**                           contain its own client-side where component.
** 010 ECF 20060424   @25670 Fix to several constructor variants. Swapped
**                           positions of sort and inverse parameters to
**                           match the natural order in which these
**                           parameters are emitted in converted code.
** 011 ECF 20060424   @25695 Added new constructor variants. Each existing
**                           variant now has a counterpart which does not
**                           expect a WhereExpression parameter.
** 012 ECF 20060424   @25699 Roll back previous change. Fixed ordering of
**                           some parameters in internal constructor
**                           calls.
** 013 ECF 20060425   @25718 Added setLenientOffEnd method. Required for
**                           conversion of certain OPEN QUERY constructs.
** 014 ECF 20060501   @25915 Added isPreselect method. Indicates query's
**                           preselection behavior.
** 015 ECF 20060531   @26850 Replaced EndConditionException with
**                           QueryOffEndException. The latter is a more
**                           specific version of the former, used to
**                           indicate a query has run off the end of its
**                           result set.
** 016 ECF 20060602   @26944 Optimization. HQL queries now will always
**                           attempt to load the full record, rather than
**                           just the primary key ID. This avoids the
**                           overhead of a separate load step.
** 017 ECF 20060616   @27412 Fix to regression caused by #016 (@26944).
**                           FilteredResults needs to store DMO instance
**                           within an array.
** 018 ECF 20060705   @27828 Fixed ClassCastException. Made fetch method
**                           tolerant of a primary key ID result (instead
**                           of DMO) in a result set.
** 019 ECF 20060713   @28051 Integrated HQLPreprocessor. An instance of
**                           this class is created for each component
**                           added to the query. This affects how the
**                           overall HQL query is composed.
** 020 ECF 20060719   @28116 Rolled back #016 (@26944). With the addition
**                           of the change broker architecture, it is
**                           actually possible to avoid loading the record
**                           at all in limited cases. In these cases, only
**                           a Hibernate proxy is created, which is far
**                           more efficient than retrieving the entire
**                           record up front. Additionally, #016 was
**                           problematic for locking, because the record
**                           could be changed by another user after it was
**                           fetched, but before the lock was acquired.
** 021 ECF 20060722   @28135 Changes to make inheritance of this class
**                           more flexible. In particular, this set of
**                           changes accommodates the AdaptiveQuery
**                           subclass.
** 022 ECF 20060725   @28188 Fixed regressions and implemented
**                           Finalizable. The latter is necessary to
**                           support retry, and to prevent a session from
**                           closing prematurely and discarding the
**                           query's results.
** 023 ECF 20060730   @28270 Client-side where clause processing changes.
**                           Moved FilteredResults into a top-level class.
** 024 ECF 20060801   @28352 Refactored execute() method and order by
**                           clause preparation code. Query substitution
**                           parameter preparation was moved into its own
**                           polymorphic method. Integrated SortCriterion
**                           class to replace code that was duplicative
**                           with that class.
** 025 ECF 20060803   @28396 Further refactoring. Added getRecordBuffers()
**                           method. Made results private again.
** 026 ECF 20060807   @28462 Fixed assembleOrderByClauseNoPreamble(). DMO
**                           alias needed to be trimmed of leading space.
** 027 ECF 20060829   @28954 Fixed join defect in constructor. We were
**                           creating a join and passing it to the wrong
**                           version of addComponent().
** 028 ECF 20060830   @28991 Additional join fix. The fix has been moved
**                           from a constructor to the most specific
**                           variant of addComponent().
** 029 ECF 20060831   @29087 Added support for sorting by an indexed
**                           property. Specifically, support sorting by
**                           a property of a DMO's associated list of
**                           composite properties.
** 030 ECF 20060901   @29139 Fixed NPE. Methods which accept a list of
**                           sort criteria must handle null list case.
** 031 ECF 20060911   @29461 Added debug code. Report DMOs found by the
**                           query if trace logging is enabled.
** 032 ECF 20060925   @29963 Check for a pending error before retrieving a
**                           record. Retrieval is not attempted if an
**                           error is pending in the ErrorManager.
** 033 ECF 20061003   @30154 Backed out #032 (@29963). This logic is only
**                           necessary for FindQuery and AdaptiveFind.
** 034 ECF 20061010   @30310 Minor change to initialization logic. Joins
**                           across different Persistence objects, rather
**                           than across databases, are now disallowed.
** 035 ECF 20061019   @30528 Fixed reposition support to better handle
**                           varying subclass behavior. This class now
**                           implements the SessionListener interface.
**                           Upon receiving a sessionClosing notification,
**                           a PreselectQuery will cache all of its
**                           results in a SimpleResults object and replace
**                           its existing results with that cache.
** 036 ECF 20061023   @30607 Moved errorIfNull logic to AbstractQuery base
**                           class. Method setErrorIfNull() is now
**                           required by the P2JQuery interface and is
**                           implemented in AbstractQuery instead of here.
** 037 ECF 20061027   @30752 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 038 ECF 20061102   @30897 Fixed issue querying row count. The cached
**                           row count must be reset to its default value
**                           when a new result set is retrieved. This
**                           forces the size() method to recalculate the
**                           proper row count.
** 039 ECF 20061102   @30959 Added support for reposition notification and
**                           auto-fetch features. Cleaned up reposition
**                           logic so that fetches after a reposition are
**                           (hopefully) correct now.
** 040 ECF 20061115   @31209 Changed call to set current buffer record.
**                           RecordBuffer API has changed to permit proper
**                           management of placeholder snapshot record.
**                           Added isResetSnapshot() method to allow
**                           subclasses to override this behavior.
** 041 ECF 20061120   @31326 Modified Component inner class. Expose
**                           restriction properties determined during HQL
**                           where clause preprocessing.
** 042 ECF 20061204   @31576 Integrated RepositionCache. This replaced a
**                           simple hash map approach which failed in
**                           cases where a lookup used an ID array which
**                           had fewer IDs than joined tables in a query.
**                           Also emulate Progress reposition quirk which
**                           repositions to last record in result set when
**                           no match is found for the target recid.
** 043 ECF 20061206   @31587 Integrated changes to AbstractQuery's
**                           afterReposition() signature. Failed attempts
**                           to reposition by ID cause afterReposition()
**                           to be invoked in notify-only mode.
** 044 ECF 20070225   @32227 Plug results resource leak. Clean up results
**                           resources when enclosing block ends. This was
**                           causing both prepared statements and result
**                           sets to remain unclosed in the backing JDBC
**                           connection.
** 045 ECF 20070225   @32232 Fixed regression caused by #044 (@32227).
**                           Finalizable which cleans up results resources
**                           must not be hard-coded to nearest enclosing
**                           scope where results are created. Subclasses
**                           may need to register this object to the
**                           global scope instead.
** 046 ECF 20070228   @32247 Optimized no-lock and temp table fetches. In
**                           these cases, it is not necessary to fetch the
**                           record in two steps. The no-lock case makes
**                           no guarantee that the record in memory has
**                           not been updated by other sessions. The temp
**                           table case guarantees that no other session
**                           can make an update to the record in memory,
**                           because a temp table record is inherently
**                           scoped to a session.
** 047 ECF 20070308   @32334 Fixed behavior for first/last. These should
**                           never raise an error condition, only a query
**                           off end condition.
** 048 ECF 20070328   @32648 Changed to support HQLPreprocessor caching.
**                           Use factory method to retrieve an instance of
**                           the preprocessor rather than invoking the
**                           constructor directly.
** 049 ECF 20070412   @32984 Accommodate method name change in
**                           RecordBuffer. setCurrentRecord --> setRecord.
** 050 ECF 20070415   @33002 Fixed locking problem. DMOs fetched "whole"
**                           with lock type NONE must be re-loaded with a
**                           lock if called for by the query component or
**                           lock type override in the retrieval method.
**                           In certain circumstances, these were being
**                           left without a lock.
** 051 ECF 20070416   @33026 Integrated user interrupt handling.
** 052 ECF 20070502   @33371 Changed to support HQLPreprocessor changes.
**                           The signature of that class' primary factory
**                           has changed.
** 053 ECF 20070504   @33409 Fixed load() implementation. We now retry
**                           reposition if it fails, assuming current
**                           result set did not contain target records.
** 054 ECF 20070508   @33459 Fixed sessionClosing() and session listener
**                           registration. Was not properly scoping
**                           registration for scrolling queries.
** 055 ECF 20070511   @33470 Fixed resetResults(). Additional state needed
**                           to be reset to default values.
** 056 ECF 20070517   @33670 Fixed abend in sessionClosing(). Check that
**                           all necessary buffers are still in scope
**                           before attempting to cache preselected
**                           results in a SimpleResults object. This
**                           prevents a fatal error when trying to
**                           retrieve results for a temp table which has
**                           been dropped by the time we get the session
**                           closing notification.
** 057 ECF 20070518   @33683 Changed access for getRecordBuffers(). Made
**                           public (was protected), since this is now a
**                           method of the Joinable interface.
** 058 ECF 20070706   @34402 Fixed selection of scope for SessionListener
**                           registration. Implemented new SessionListener
**                           method deregisteredSessionListener(). Also
**                           replaced StringBuffer with StringBuilder.
** 059 ECF 20070709   @34422 Fixed sessionClosing(). Was unsafe for
**                           certain non-scrolling results which do not
**                           support reset.
** 060 ECF 20070711   @34448 Extracted ScrollingResults inner class into a
**                           separate, top-level class.
** 061 ECF 20070716   @34510 Implement new cleanup() method as required by
**                           P2JQuery interface. Check whether query is
**                           standalone before registering a cleaner with
**                           the TM. This prevents us from cleaning up too
**                           aggressively in case we are wrapped by a
**                           QueryWrapper or contained in a CompoundQuery
**                           that is so wrapped.
** 062 ECF 20070720   @34611 Fixed fetch(). Always reload a DMO, even if
**                           it is in a temp table or has no lock. This is
**                           necessary to handle the case of fetching a
**                           cached DMO which may have been detached from
**                           its session previously. For a persistent DMO,
**                           the additional overhead should be small,
**                           since Hibernate only needs to go to the first
**                           level cache to find the object.
** 063 ECF 20070720   @34621 Implemented a better fix for fetch() than
**                           #062 (@34611). If DMO is a temp table, simply
**                           try to reassociate it with the Hibernate
**                           session, rather than going all the way back
**                           to the database (if DMO was detached). Can't
**                           do this for permanent tables, though.
** 064 ECF 20070821   @34907 Added support for temporary silent error
**                           mode. This allows query substitution parms to
**                           be processed safely during initialization of
**                           a query.
** 065 ECF 20070827   @34996 Optimized certain temp table queries. Only
**                           augment where clause with a temp buffer's
**                           multiplex ID when necessary. Use multiplex ID
**                           as a constant in where clause rather than as
**                           a substitution parameter.
** 066 ECF 20070927   @35276 Fixed reposition logic. Any reposition must
**                           first release all buffers. Changes to methods
**                           to get current row and report whether query
**                           is off-end.
** 067 ECF 20071030   @35634 Fixed fetch() to prevent loading stale temp
**                           table records. It is possible for a record to
**                           be evicted from the session cache while still
**                           held in a result set, such that the result
**                           set will return a stale record that is not a
**                           detached instance, even if the session is
**                           flushed before the fetch. We now force a
**                           reload of fetched temp table records to avoid
**                           this problem.
** 068 ECF 20071130   @36122 Removed isRefreshSnapshot() method. The
**                           RecordBuffer placeholder snapshot is no
**                           longer cleared by any query type when the
**                           query does not find a record to place in the
**                           buffer. Integrated generics.
** 069 ECF 20071206   @36267 Fixed off-end tracking. Replaced boolean flag
**                           with OffEnd enum to discern between off-end
**                           directions (i.e., off back vs off front of
**                           result set). Current off-end status and
**                           active sort index are reported to the record
**                           buffer when fetching a record.
** 070 ECF 20071218   @36463 Fixed order by clause generation. Was failing
**                           for sort criteria which included references
**                           to specific extents of indexed properties.
** 071 ECF 20080120   @37009 Removed restriction of dealing only with IDs.
**                           This query can now handle both Persistables
**                           and primary key IDs when retrieving data.
**                           This is necessary when the full record is
**                           needed immediately, not just an ID, such as
**                           in presort cases. Also added open() method to
**                           set proper semantics for a converted OPEN
**                           QUERY statement, and optionally to reposition
**                           to the first result.
** 072 ECF 20080222   @37144 Fixed NPE in sessionClosing().
** 073 ECF 20080310   @37481 Made changes necessary to support embedded,
**                           H2 database.
** 074 SVL 20080418   @38036 Added support for the inverseSorting flag.
** 075 SVL 20080421   @38080 Added support for legacy key join helpers.
** 076 SVL 20080425   @38119 lastRepo assignment has been fixed into
**                           reposition(int).
** 077 ECF 20080502   @23396 Backed out #076 (@38119). The reposition(int)
**                           implementation was generally correct and this
**                           change caused regressions. Fixed open() to
**                           execute the query immediately instead of
**                           doing this lazily. We now reposition on open
**                           only if the query is being browsed. Removed
**                           obsolete method isRepositionOnOpen().
** 078 SVL 20080508   @38238 Fixed incorrect record values after
**                           first-next, forward/backward(0) and
**                           last-prev execution sequences.
** 079 ECF 20080508   @38243 Fixed regression in reposition(int) which did
**                           not handle all non-positive rows correctly.
**                           Refactored open() method to use parent class.
**                           Override base lock type for temporary tables.
**                           This allows some downstream optimizations.
** 080 ECF 20080509   @38252 Fixed regression in open(). Removed check for
**                           standalone flag.
** 081 SVL 20080526   @38311 Preselect locking implemented.
** 082 SVL 20080604   @38534 Modified according to the new P2JQuery
**                           interface. previous() fixed.
** 083 ECF 20080604   @38608 Added placeholder for dirty database check.
** 084 SVL 20080611   @38685 Behavior of previous() fixed for the case 
**                           when the off-end is FRONT.
** 085 CA  20080618   @38869 When the JOINed foreign buffer does not refer
**                           any record, the query must return no results
**                           as it can not be executed.
** 086 CA  20080619   @38871 If executed within an OPEN QUERY statement, 
**                           the query must be removed from the session's
**                           listeners only when the enclosed external
**                           scope is exit. This fixes the case when the
**                           OPEN QUERY is executed inside a procedure.
** 087 ECF 20080709   @39086 Created cacheResults() method. Abstracted
**                           from sessionClosing() and made protected so
**                           subclasses could access this functionality.
** 088 ECF 20080710   @39090 Modified sessionClosing() and cacheResults().
**                           The latter no longer throws
**                           IllegalStateException.
** 089 ECF 20080714   @39127 Fixed NPE in debug logging.
** 090 ECF 20080723   @39173 Removed retry handling. This is now the
**                           responsibility of subclasses.
** 091 CA  20080729   @39218 Added back the retry handling. Now the retry
**                           will rely on the reposition mechanism to 
**                           retrieve the same record, but only when the
**                           query is associated with a FOR block.
** 092 SVL 20080730   @39237 Component.multiplexID field was removed.
**                           RecordBuffer's multiplex ID is used instead.
** 093 ECF 20080730   @39325 Added rowid support. Refactored reposition*()
**                           methods.
** 094 SVL 20080808   @39348 Fixed cacheResults(). Did not handle case
**                           when query has just executed but has not yet
**                           retrieved any results.
** 095 ECF 20080822   @39556 Fix for join validation. Abstracted check
**                           into separate method, verifyJoins(). Disable
**                           join buffer check for server-based joins.
**                           Other changes in support of global DMO change
**                           events.
** 096 CA  20080819   @39463 Support API change in ScrollingResults.
** 097 CA  20080828   @39610 Fix NPE caused by repositionByID(rowid, ...).
** 098 ECF 20081009   @40081 Added support for inlining query substitution
**                           parameters. Integrated HQLPreprocessor API
**                           changes and the ParameterIndices class.
** 099 ECF 20081030   @40308 Fixed regression introduced by #098 (@40081).
**                           Refactored Component inner class and code in
**                           top-level class which uses it.
** 100 ECF 20081105   @40311 Implemented registerRecordChangeListeners().
**                           Required by the Joinable interface.
** 101 ECF 20081106   @40324 Added _isOffEnd(). Required by P2JQuery
**                           interface change.
** 102 ECF 20081107   @40405 Moved multiplex ID position in order by
**                           clause. It used to be the first element, but
**                           that sometimes resulted in a less efficient
**                           index being selected for a query plan. Now it
**                           is either the last element, or second to last
**                           (if the primary key is part of the clause).
** 103 CA  20081210   @40839 Do not cleanup if the query was explicitly 
**                           closed.
** 104 CA  20081211   @40859 Fixed regression introduced by H103: on 
**                           sessionClosing event, if the query is closed,
**                           return true.
** 105 ECF 20081215   @40914 Removed flushBuffers() method. This work is
**                           now done when a component is added to the
**                           query. The previous invocation, during record
**                           retrieval, was too late. A validation error
**                           ignores silent error mode, since the flushing
**                           of the previous record in the buffer is not
**                           masked by silent error mode.
** 106 ECF 20090202   @41248 Rolled back #081 (@38311). Prelock mechanism
**                           has been disabled.
** 107 ECF 20090313   @41600 Changed createResults(). Added parameter to
**                           FilteredResults constructor.
** 108 SVL 20090316   @41610 assembleOrderByClauseNoPreamble() refactored,
**                           sort variable is protected instead of private 
**                           now.
** 109 GES 20090424   @41941 Import change.
** 110 ECF 20090512   @42153 Modified addComponent(). Now invokes flushAll()
**                           instead of flush() on backing buffer. This
**                           ensures all like buffers are allowed to flush
**                           their transient records properly.
** 111 ECF 20090603   @42599 Fixed fetch(). A dirty record in the backing
**                           buffer must not be replaced by the primary
**                           database's version of that record.
** 112 ECF 20090604   @42606 Added support for Progress isolation leak quirk.
**                           Modified execute() method to publish uncommitted
**                           changes with dirty share manager if necessary.
** 113 ECF 20090609   @42631 Fixed execute(). We now evict, if possible, those
**                           DMOs read into the Hibernate Session, whose
**                           uncommitted changes are leaked to other sessions
**                           early.
** 114 ECF 20090619   @42772 Added executeScroll(). Executes default variant
**                           of Persistence.scroll(String, Object[], Type[]).
**                           Children classes which need to invoke different
**                           variants must override this method.
** 115 ECF 20090630   @42985 Rewrote reposition logic. Moved from integer to
**                           recid type in reposition methods.
** 116 ECF 20090702   @43035 Fixed sessionClosing() and regressions in query
**                           (in the Progress OPEN QUERY sense) behavior.
**                           Added isNativeScrolling() to indicate that a base
**                           PreselectQuery instance used for a converted 4GL
**                           query behaves as if that query were defined as
**                           scrolling. Added error handling in caching
**                           results during a sessionClosing event.
** 117 ECF 20090703   @43051 Reimplemented load() method. This method now sets
**                           the backing buffer to unknown mode and throws
**                           MissingRecordException if a target record cannot
**                           be loaded. Also fixed cleanup(); the query is now
**                           flagged as closed once it runs.
** 118 ECF 20090707   @43101 Fixed defect related to processing reposition
**                           notifications. Make sure we don't release the
**                           buffers' current records and change other state
**                           when processing nested reposition calls during
**                           reposition notifications.
** 119 ECF 20090708   @43107 Partial rollback of #118 (@43101). Some state
**                           must be changed when processing nested reposition
**                           calls during reposition notifications.
** 120 ECF 20090711   @43192 Use new ScrollingResults c'tor.
** 121 ECF 20090716   @43220 Refactored for lazy record buffer initialization.
**                           Initialize backing buffers when their scopes have
**                           been opened. Otherwise allow initialization to be
**                           deferred.
** 122 SVL 20090723   @43344 Added setNonScrolling() function.
** 123 ECF 20090724   @43419 Made getPersistence() protected.
** 124 ECF 20090729   @43448 Fixed executeQuery(). Surrounded creation of
**                           ScrollingResults with push/pop calls for implicit
**                           transaction.
** 125 CA  20090731   @43468 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 126 ECF 20090810   @43588 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 127 ECF 20090811   @43600 Fixed resource cleanup. Refactored registration
**                           of a cleaner object with the TM, so that
**                           subclasses could use it. Ensured cleaner is only
**                           registered once.
** 128 SVL 20091030   @44296 Fixed forward(), backward(), currentRow() and next() functions.
** 129 OM  20130829          Added change detection to buffer on current() methods.
**                           Updated calls to match the new Persistence.load() signatures.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Query which issues an HQL query to the database server and uses a
 * scrollable database cursor to navigate over the set of results.  This
 * query is used to back converted Progress queries which either explicitly
 * or implicitly cause a result list to be preselected.  Both single- and
 * multi-table queries are supported.  Because this class implements the
 * <code>Joinable</code> interface, instances may be added as components of
 * a {@link CompoundQuery}.
 * <p>
 * The query is lazily executed the first time a record retrieval/navigation
 * method is invoked.  Only the database identifiers (i.e., primary keys) of
 * the records are retrieved by the initial query.  On each subsequent fetch,
 * a row in this result list is retrieved, where each column contains the
 * primary key for the record associated with the corresponding query
 * component.  The record for each primary key is retrieved and stored in its
 * associated buffer.
 * <p>
 * At this time, outer joins are not supported.  Furthermore, components may
 * only specify NEXT as their iteration type at this time.  FIRST and LAST
 * currently are not supported.
 * <p>
 * When used to represent a Progress query (created with DEFINE or OPEN
 * QUERY), lenient-off-end must be enabled at construction.  This mode
 * suppresses end condition exceptions for <code>next</code> and
 * <code>previous</code> commands which run off the end of the query's
 * results.
 * <p>
 * This query may be integrated with a client-side where clause expression,
 * by passing a non-<code>null</code> <code>WhereExpression</code> argument
 * to one of the single-table constructor variants.  Note that additional
 * components may not be added to an instance so created, since this type of
 * query must act against no more than one record buffer.  If a multi-table
 * preselect query is called for, <code>CompoundQuery</code> must be used,
 * though its components may be instances of this class.
 */
public class PreselectQuery
extends AbstractQuery
implements Joinable,
           QueryConstants,
           Finalizable,
           SessionListener
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(PreselectQuery.class);
   
   /** Has query been repositioned between result rows? */
   protected boolean betweenRows = true;
   
   /** Has the query run off either end of its result set? */
   protected OffEnd offEnd = OffEnd.NONE;
   
   /** HQL order by clause used across all query components */
   protected final String sort;

   /** Components of the query;  at least 1 is required */
   private final List<Component> components;
   
   /** List of client-side where clause expressions */
   private final List<WhereExpression> whereExprList =
      new ArrayList<WhereExpression>(1);
   
   /** Persistence object which this query's components use */
   private Persistence persistence = null;
   
   /** Flag indicating results must be scrollable */
   private boolean scrolling = false;
   
   /** Cache of primary key lists to result set row numbers */
   private RepositionCache repoCache = new RepositionCache();
   
   /** HQL query string to retrieve primary key ID arrays */
   private String hql = null;
   
   /** Result set for this query */
   private Results results = null;
   
   /** Does this query have any arguments which are runtime resolvable? */
   private boolean resolvableArgs = false;
   
   /** <code>false</code> if running off end of results should raise error */
   private boolean lenientOffEnd = false;
   
   /** Number of total results returned */
   private int rowCount = -1;
   
   /** Is query registered for session event notification? */
   private boolean regWithSession = false;
   
   /** Has query been registered for resource cleanup? */
   private boolean regCleaner = false;
   
   /** Force query to retrieve full records, rather than primary keys only */
   private boolean fullRecords = false;
   
   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    *
    * @param   sort
    *          HQL order by clause to be applied across all components of the
    *          query.
    */
   public PreselectQuery(String sort)
   {
      this.sort = sort;
      components = new ArrayList<Component>(1);

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * This variant is used by code which is converted from an OPEN QUERY
    * statement, which permits advancing off the end of a result set without
    * raising a <code>QueryOffEndException</code>.
    *
    * @param   lenientOffEnd
    *          <code>true</code> if advancing past either end of the result
    *          set should NOT raise a <code>QueryOffEndException</code>;
    *          <code>false</code> if doing so should raise the exception.
    * @param   sort
    *          HQL order by clause to be applied across all components of the
    *          query.
    */
   public PreselectQuery(boolean lenientOffEnd, String sort)
   {
      this(sort);
      this.lenientOffEnd = lenientOffEnd;
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort)
   {
      this(dmo, where, whereExpr, sort, null, null, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         DataModelObject inverse)
   {
      this(dmo, where, whereExpr, sort, inverse, null, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         Object[] args)
   {
      this(dmo, where, whereExpr, sort, null, args, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         DataModelObject inverse,
                         Object[] args)
   {
      this(dmo, where, whereExpr, sort, inverse, args, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         LockType lockType)
   {
      this(dmo, where, whereExpr, sort, null, null, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         DataModelObject inverse,
                         LockType lockType)
   {
      this(dmo, where, whereExpr, sort, inverse, null, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         Object[] args,
                         LockType lockType)
   {
      this(dmo, where, whereExpr, sort, null, args, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PreselectQuery(DataModelObject dmo,
                         String where,
                         WhereExpression whereExpr,
                         String sort,
                         DataModelObject inverse,
                         Object[] args,
                         LockType lockType)
   {
      super();
      
      components = new ArrayList<Component>(2);
      
      setErrorIfNull(false);
      addComponent(dmo, where, inverse, args, lockType, NEXT, false);
      
      this.sort = sort;
      if (whereExpr != null)
      {
         addWhereExpression(whereExpr);
      }

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Set the behavior of this query as a record retrieval request moves off
    * the end of its results.  If set to lenient mode, invocations of
    * <code>next</code> and <code>previous</code> which do not find a record
    * do not raise a <code>QueryOffEndException</code>.  By default, this
    * would normally raise the exception.
    *
    * @param   lenientOffEnd
    *          <code>true</code> to suppress <code>QueryOffEndException</code>
    *          for <code>next</code> and <code>previous</code> requests;
    *          <code>false</code> to permit these exceptions.
    */
   public void setLenientOffEnd(boolean lenientOffEnd)
   {
      this.lenientOffEnd = lenientOffEnd;
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo, String where)
   {
      addComponent(dmo, where, null, null, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse)
   {
      addComponent(dmo, where, inverse, null, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo, String where, Object[] args)
   {
      addComponent(dmo, where, null, args, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args)
   {
      addComponent(dmo, where, inverse, args, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            LockType lockType)
   {
      addComponent(dmo, where, null, null, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            LockType lockType)
   {
      addComponent(dmo, where, inverse, null, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, null, args, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, inverse, args, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      Component comp = new Component((BufferReference) dmo,
                                     null,
                                     where,
                                     lockType,
                                     args,
                                     iteration,
                                     outer);
      addComponent(comp);
   }
   
   /**
    * Add a component to the query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      BufferReference proxy = (BufferReference) dmo;
      AbstractJoin join = processJoin(proxy, (BufferReference) inverse);
      
      Component comp = new Component(proxy,
                                     join,
                                     where,
                                     lockType,
                                     args,
                                     iteration,
                                     outer);
      addComponent(comp);
   }
   
   /**
    * Register all <code>RecordChangeListener</code>s associated with this
    * query (if any), with the context-local <code>ChangeBroker</code> at the
    * indicated scope.
    * <p>
    * This implementation does nothing.
    * 
    * @param   scope
    *          Scope at which listeners should be registered with the change
    *          broker.
    */
   public void registerRecordChangeListeners(int scope)
   {
   }
   
   /**
    * Respond to a clean (i.e., transaction commit) session event by caching
    * all preselected results and using that cache to back the query's results
    * for all future fetches.  If all results already are cached, do nothing.
    * <p>
    * Respond to a session closing event by clearing all results, unless they
    * were just cached as described above.
    *
    * @param   event
    *          Session event type.
    *
    * @return  <code>true</code>, indicating this query should be
    *          deregistered as a session listener after this method returns;
    *          else <code>false</code>.
    * 
    * @throws  PersistenceException
    *          if there is an error caching results.
    */
   public boolean sessionEvent(SessionListener.Event event)
   throws PersistenceException
   {
      if (closed)
      {
         return true;
      }
      
      switch (event)
      {
         case TRANSACTION_COMMITTING:
            if (results != null)
            {
               try
               {
                  SimpleResults cached = cacheResults(results);
                  if (results != cached)
                  {
                     results.sessionClosing();
                  }
                  setResults(cached);
               }
               catch (Exception exc)
               {
                  resetResults();
                  
                  throw new PersistenceException("Error caching results",
                                                 exc);
               }
            }
            break;
            
         case SESSION_CLOSING_NORMALLY:
            if (results != null && !(results instanceof SimpleResults))
            {
               resetResults();
            }
            break;
            
         case SESSION_CLOSING_WITH_ERROR:
            resetResults();
            break;
            
         default:
            return false;
      }
      
      return true;
   }
   
   /**
    * Invoked when this session listener is removed from the list of
    * registered session listeners.
    */
   public void deregisteredSessionListener()
   {
      regWithSession = false;
   }
   
   /**
    * Provides a notification that the scope in which this query was created
    * is ending.  This is a no-op implementation required by the interface.
    */
   public void finished()
   {
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to iterate and attempt another pass.  This is a
    * no-op implementation required by the interface.
    */
   public void iterate()
   {
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to retry the same iteration (all loop control
    * data is unchanged).  This implementation sets the state of the query as
    * if a reposition to the current row has just occurred.  This will keep
    * the subsequent call to {@link #next()} at the top of the block body from
    * advancing to the next row.
    * <p>
    * This implementation retries the current record only if the query is 
    * associated with a FOR block (see {@link BlockType}).
    */
   public void retry()
   {
      BlockType type = TransactionManager.getBlockType();
      
      // only queries associated with a FOR block allows the retry of the 
      // current record
      switch (type)
      {
         case FOR_LOOP:
         case FOR_LOOP_WHILE:
         case FOR_LOOP_TO:
         case FOR_LOOP_TO_WHILE:
         case FOR_BLOCK:
         case FOR_BLOCK_WHILE:
         case FOR_BLOCK_TO:
         case FOR_BLOCK_TO_WHILE:
            betweenRows = true;            
            break;
      }

      if (LOG.isDebugEnabled())
      {
         LOG.debug("Preparing query for retry [" + this + "]");
      }
   }
   
   /**
    * Create a string representation of this object's state, primarily for
    * debug purposes.
    * 
    * @return  String representation of this object.
    */
   public String toString()
   {
      StringBuilder bld = new StringBuilder();
      bld.append("[");
      bld.append(getTableCount());
      bld.append(" table(s)] HQL: ");
      bld.append(hql);

      return bld.toString();
   }
   
   /**
    * Add a client-side where expression to this query.  Each initial query
    * result found by the server must be accepted by this expression before
    * becoming a part of the query's final result set.
    *
    * @param   whereExpr
    *          Client-side where clause expression.  May not be
    *          <code>null</code>.
    *
    * @throws  NullPointerException
    *          if <code>whereExpr</code> is <code>null</code>.
    */
   public void addWhereExpression(WhereExpression whereExpr)
   {
      if (whereExpr == null)
      {
         throw new NullPointerException(
            "Client-side where expression may not be null");
      }
      
      whereExprList.add(whereExpr);
      whereExpr.resolveArgs();
   }
   
   /**
    * Set the iteration type for the most recently added query component.
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}.
    *
    * @throws  IllegalStateException
    *          If no component has yet been added.
    */
   public void setIterationType(int iteration)
   {
      if (components.isEmpty())
      {
         throw new IllegalStateException(
            "Must add a query component before setting its iteration type");
      }
      
      Component comp = components.get(components.size() - 1);
      comp.setIteration(iteration);
   }
   
   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first()
   {
      first(null);
   }
   
   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first(LockType lockType)
   {
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = (missing ? results.next() : results.first());
            missing = false;
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            betweenRows = false;
            fetch(avail, lockType, isErrorIfNull());
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            exc.fillInStackTrace();
            throw exc;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the last composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void last()
   {
      last(null);
   }
   
   /**
    * Retrieve the last composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void last(LockType lockType)
   {
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = (missing ? results.previous() : results.last());
            missing = false;
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            betweenRows = false;
            fetch(avail, lockType, isErrorIfNull());
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            exc.fillInStackTrace();
            throw exc;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next()
   {
      next(null);
   }
   
   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next(LockType lockType)
   {
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = ((betweenRows && results.getRowNumber() != -1) || results.next());
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            fetch(avail, lockType, false);
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            if (!lenientOffEnd)
            {
               exc.fillInStackTrace();
               throw exc;
            }
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         finally
         {
            betweenRows = false;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the previous composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void previous()
   {
      previous(null);
   }
   
   /**
    * Retrieve the previous composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void previous(LockType lockType)
   {
      boolean setToLast = treatPreviousAsLast();
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = (setToLast ? results.last() : results.previous());
            if (avail)
            {
               offEnd = OffEnd.NONE;
            }
            else
            {
               offEnd = (setToLast ? OffEnd.BACK : OffEnd.FRONT);
            }
            fetch(avail, lockType, false);
         }
         catch (QueryOffEndException exc)
         {
            offEnd = (setToLast ? OffEnd.BACK : OffEnd.FRONT);
            if (!lenientOffEnd)
            {
               exc.fillInStackTrace();
               throw exc;
            }
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         finally
         {
            betweenRows = false;
         }
      } while (missing);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query).
    *
    * @throws  ErrorConditionException
    *          if the query fails or if more than one result is found which
    *          matches the query criteria.
    * @throws  UnsupportedOperationException
    *          if this method is invoked on a multi-table query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void unique()
   {
      unique(null);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query),
    * overriding the default lock type of the underlying query component.
    *
    * @param   lockType
    *          Lock type which should override that of the query component.
    *
    * @throws  ErrorConditionException
    *          if the query fails or if more than one result is found which
    *          matches the query criteria.
    * @throws  UnsupportedOperationException
    *          if this method is invoked on a multi-table query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void unique(LockType lockType)
   {
      if (components.size() > 1)
      {
         throw new UnsupportedOperationException(
            "Unique retrieval only valid for a single-table preselect query");
      }
      
      boolean setToFirst = (results == null);
      execute();
      boolean avail = true;
      if (setToFirst)
      {
         avail = results.first();
      }
      
      if (!results.isFirst() || !results.isLast())
      {
         try
         {
            Component comp = components.get(0);
            RecordBuffer buffer = comp.getBuffer();
            getPersistence().uniqueResultViolation(buffer, null);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
            
            return;
         }
      }
      
      offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
      
      try
      {
         fetch(avail, lockType, false);
      }
      catch (MissingRecordException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      catch (QueryOffEndException exc)
      {
         offEnd = OffEnd.BACK;
         exc.fillInStackTrace();
         throw exc;
      }
   }
   
   /**
    * Reload the current composite row of results for the query.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the reload triggers an error.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void current()
   {
      current(null);
   }
   
   /**
    * Reload the current composite row of results for the query, overriding
    * the lock type to apply to each record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any query component.
    *
    * @throws  ErrorConditionException
    *          if the reload triggers an error.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void current(LockType lockType)
   {
      execute();
      
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component comp = iter.next();
         LockType lt = (lockType == null ? comp.getLockType() : lockType);
         
         final RecordBuffer buffer = comp.getBuffer();
         buffer.armCurrentChanged();
         buffer.reload(lt);
         buffer.updateCurrentChanged();
      }
   }
   
   /**
    * Load one or more records into their associated buffers, given an array
    * of primary key ID values, or the actual DMOs themselves.  The IDs/DMOs
    * are arranged to coincide with the tables underlying the query, from
    * left-most to right-most (in the sense of how they are joined by the
    * query).
    * <p>
    * In the event an expected value cannot be loaded (e.g., the record has
    * been deleted or is otherwise no longer available), this method throws
    * {@link MissingRecordException}, after setting the associated buffer(s)
    * into {@link RecordBuffer#setUnknownMode() unknown mode}. Note that a
    * record being locked by another session does not constitute a "missing"
    * record. Note also that this exception is not thrown until all the
    * elements of <code>data</code> have been processed to the extent
    * possible.
    *
    * @param   data
    *          Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    * @param   lockType
    *          Lock type which should by used.  Set to <code>null</code> to
    *          allow query to use its current lock type.
    * 
    * @throws  MissingRecordException
    *          if any record cannot be loaded, because it is no longer
    *          available (e.g., deleted, etc.).
    * @throws  PersistenceException
    *          if there is an error loading data.
    */
   public void load(Object[] data, LockType lockType)
   throws MissingRecordException,
          PersistenceException
   {
      boolean retry = (results != null);
      
      int len = data.length;
      Serializable[] ids = new Serializable[len];
      for (int i = 0; i < len; i++)
      {
         Object datum = data[i];
         ids[i] = (datum instanceof Persistable
                   ? ((Persistable) datum).getId()
                   : (Serializable) datum);
      }
      
      try
      {
         if (!repositionByID(ids, true))
         {
            if (retry)
            {
               // Query did not have the target record(s) cached in its current
               // result set, so reset the results and try again.
               resetResults();
               if (!repositionByID(ids, true))
               {
                  throw new MissingRecordException();
               }
            }
            else
            {
               throw new MissingRecordException();
            }
         }
      
         fetch(offEnd == OffEnd.NONE, lockType, false);
      }
      catch (MissingRecordException exc)
      {
         setEmptyBuffersUnknown();
         
         throw exc;
      }
      finally
      {
         betweenRows = false;
      }
   }
   
   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      List<QueryOffEndListener> listeners = 
         new ArrayList<QueryOffEndListener>();
      
      for (int i = 0; i < components.size(); i++)
      {
         listeners.add(components.get(i).buffer.getQueryOffEndListener()); 
      }
      
      return listeners;
   }

   /**
    * Assemble an array of primary key IDs for the current record(s) in the
    * buffers underlying this query, from left-most to right-most (in the
    * sense of how the query joins the associated tables).  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * record(s), so that this may be restored later.
    *
    * @return  Array of primary key IDs, one per table involved in the query.
    *
    * @see     #load(Object[], LockType)
    */
   public Object[] getRow()
   {
      if (results == null)
      {
         return null;
      }
      
      int size = components.size();
      Object[] rowData = new Object[size];
      System.arraycopy(results.get(), 0, rowData, 0, size);
      
      return rowData;
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   public void repositionByID(recid id)
   {
      if (!validateRepositionByID(id))
      {
         return;
      }
      
      Serializable[] serIDs = new Serializable[] { new Long(id.longValue()) };
      
      if (!repositionByID(serIDs, false))
      {
         ErrorManager.recordOrThrowError(
            7331,
            "Cannot reposition query to recid/rowid(s) given");
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    * 
    * @throws  ErrorConditionException
    *          if any of the specified IDs represents the unknown value.
    */
   public void repositionByID(rowid id1, rowid...joinIDs)
   {
      if (!validateRepositionByID(id1))
      {
         return;
      }
      
      int len = joinIDs.length;
      Serializable[] serIDs = new Serializable[len + 1];

      serIDs[0] = id1.getValue();
      
      for (int i = 0; i < len; i++)
      {
         rowid next = joinIDs[i];
         
         if (!validateRepositionByID(next))
         {
            return;
         }
         
         serIDs[i + 1] = next.getValue();
      }
      
      if (!repositionByID(serIDs, false))
      {
         ErrorManager.recordOrThrowError(
            7331,
            "Cannot reposition query to recid/rowid(s) given");
      }
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if <code>row</code> represents unknown value.
    */
   public void reposition(NumberType row)
   {
      if (validateReposition(row))
      {
         reposition(row.intValue());
      }
   }
   
   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void reposition(int row)
   {
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }
      
      row--;   // adjust for 1-based row
      int currentRow = results.getRowNumber();
      if ((row >= 0 && currentRow == row) || results.setRowNumber(row))
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = (row < 0 ? OffEnd.FRONT : OffEnd.BACK);
         betweenRows = false;
      }

      afterReposition(true);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if <code>rows</code> represents the unknown value;
    *          if there is an error repositioning the cursor.
    */
   public void forward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         forward(rows.intValue());
      }
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void forward(int rows)
   {
      if (rows == 0)
      {
         return;
      }

      if (rows < 0)
      {
         backward(-rows);
         return;
      }

      prepareReposition();
      
      if (results == null)
      {
         execute();
      }

      // Increment the requested number of rows by one if we are:
      // 1. on a record (i.e. not between rows) OR
      // 2. into front offend,
      // since the resulting position must be before the record *following*
      // the targeted jump.
      if (!betweenRows ||
          (results.getRowNumber() == -1 && offEnd != OffEnd.BACK))
      {
         rows++;
      }

      boolean avail = results.scroll(rows);
      if (avail)
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = OffEnd.BACK;
         betweenRows = false;
      }
      
      afterReposition(true);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if <code>rows</code> represents the unknown value;
    *          if there is an error repositioning the cursor.
    */
   public void backward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         backward(rows.intValue());
      }
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void backward(int rows)
   {
      if (rows == 0)
      {
         return;
      }

      if (rows < 0)
      {
         forward(-rows);
         return;
      }

      prepareReposition();
      
      if (results == null)
      {
         execute();
      }

      // Decrement the requested number of rows by one if we are:
      // 1. on a record (i.e. not between rows) AND
      // 2. NOT into back offend,
      // since the resulting position must be before the record *following*
      // the targeted jump.
      if (!betweenRows && offEnd != OffEnd.BACK)
      {
         rows--;
      }
      
      boolean avail = results.scroll(-rows);
      if (avail)
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = OffEnd.FRONT;
         betweenRows = false;
      }

      afterReposition(true);
   }
   
   /**
    * Return the number of results found for this query.
    * <p>
    * Note that the first time this method is invoked, the database server
    * may have to scroll to the end of the result set, which may be a
    * time-consuming process for a large result set.
    *
    * @return  Size of result set or value if the query is not open.
    */
   public integer size()
   {
      if (results == null)
      {
         return new integer();
      }
      
      if (rowCount < 0)
      {
         int row = results.getRowNumber();
         results.last();
         rowCount = results.getRowNumber() + 1;
         
         if (row == -1)
         {
            if (offEnd == OffEnd.BACK)
            {
               results.next();
            }
            else
            {
               results.reset();
            }
         }
         else
         {
            results.setRowNumber(row);
         }
      }
      
      return (new integer(rowCount));
   }
   
   /**
    * Progress-compatible function that returns the 1-based index of the
    * current row in the result set.
    *
    * @return  Unknown value if the query is not open or result set is empty.
    *          1 if the query is positioned before the first record.
    *          A value 1 greater than the number of rows in the query result
    *          set if the query is positioned beyond the last record.
    *          Otherwise returns the number of rows in the query result set.
    *
    */
   public integer currentRow()
   {
      if (results == null ||
          (offEnd == OffEnd.NONE &&
           results.getRowNumber() == -1 &&
           !isScrolling()))
      {
         return (new integer());
      }
      else
      {
         int size = size().intValue();
         int row = results.getRowNumber();

         if (size == 0)
         {
            return (new integer());
         }
         else if (offEnd == OffEnd.BACK)
         {
            return (new integer(size + 1));
         }
         else if (row == -1)
         {
            // We are at the front offend. Return unknown value if the query
            // is an adaptive one and we haven't retrieved any results using 
            // it.
            return !isNativelyPreselect() && (offEnd != OffEnd.FRONT) ?
                   new integer() :
                   new integer(1);
         }
         else
         {
            return (new integer(row + 1));
         }
      }
   }

   /**
    * Get the 1-based index of the current row in the result set.
    *
    * @return  Current row or the unknown value if the query has not yet
    *          executed.
    */
   public integer currentRowImpl()
   {
      if (results == null || offEnd != OffEnd.NONE)
      {
         return (new integer());
      }
      
      return (new integer(results.getRowNumber() + 1));
   }
   
   /**
    * Indicate whether the query has run off either end of its result set,
    * due to scrolling beyond the last record or before the first record.
    *
    * @return  <code>true</code> if no record can be retrieved at the
    *          cursor's current position;  otherwise <code>false</code>.
    */
   public logical isOffEnd()
   {
      boolean isOffEnd = (offEnd != OffEnd.NONE);
      if (!isOffEnd && results == null)
      {
         return (new logical());
      }
      
      return (new logical(isOffEnd));
   }
   
   /**
    * Indicate whether the query has run off either end of its result set,
    * due to scrolling beyond the last record or before the first record.
    *
    * @return  <code>true</code> if no record can be retrieved at the
    *          cursor's current position;  otherwise <code>false</code>.
    */
   public boolean _isOffEnd()
   {
      boolean isOffEnd = (offEnd != OffEnd.NONE);
      if (!isOffEnd && results == null)
      {
         return false;
      }
      
      return isOffEnd;
   }

   /**
    * Force the query to retrieve full records, rather than primary keys only.
    */
   public void setFullRecords()
   {
      this.fullRecords = true;
   }
   
   /**
    * Reset the query by clearing its state to a known default.  This is
    * used to ensure that subsequent requests for FIRST and LAST record
    * retrievals operate on a known query/buffer state.
    * <p>
    * This implementation nulls out the current result set, so that the next
    * record retrieval request triggers a new server-side query.  Optionally,
    * it also resets resolvable substitution arguments and triggers where
    * expression arguments to be resolved, if a client-side where cause
    * expression is defined.
    *
    * @param   resolveArgs
    *          <code>true</code> if query substitution arguments should be
    *          resolved/reset as part of the reset;  <code>false</code> if
    *          existing argument values should be used for the next query
    *          execution.
    */
   public void reset(boolean resolveArgs)
   {
      resetResults();
      
      if (resolveArgs && resolvableArgs)
      {
         resolveClientWhereArgs();
         
         for (Component comp : components)
         {
            comp.resetArgs();
         }
      }
   }
   
   /**
    * Set each buffer backing a query component to unknown mode.
    *
    * @see     RecordBuffer#setUnknownMode
    */
   public void setUnknownRecord()
   {
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component comp = iter.next();
         comp.getBuffer().setUnknownMode();
      }
   }
   
   /**
    * Get the number of tables joined by this query.
    *
    * @return  Number of joined tables.
    */
   public int getTableCount()
   {
      return components.size();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the first result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the first query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekFirst()
   {
      execute();
      results.first();
      
      return getRow();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the last result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the last query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekLast()
   {
      execute();
      results.last();
      
      return getRow();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the next result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the next query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekNext()
   {
      execute();
      results.next();
      
      return getRow();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the previous result
    * in this query's result list.
    *
    * @return  Array of record IDs reflecting the previous query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekPrevious()
   {
      execute();
      results.previous();
      
      return getRow();
   }
   
   /**
    * Indicate whether this query preselects all of its results.
    *
    * @return  <code>true</code>.
    */
   public boolean isPreselect()
   {
      return true;
   }

   /**
    * Set the given results object into this query.
    *
    * @param   results
    *          New result set for this query.
    */
   public void setResults(Results results)
   {
      this.results = results;
   }
   
   /**
    * Access the record buffers manipulated by this query, in order from
    * left-most (outermost) table of the query's join (if any) to right-most
    * (innermost) table.
    * 
    * @return  Array of one or more record buffers.
    */
   public RecordBuffer[] getRecordBuffers()
   {
      RecordBuffer[] buffers = new RecordBuffer[getTableCount()];
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         buffers[i] = iter.next().getBuffer();
      }
      
      return buffers;
   }
   
   /**
    * Clean up query resources when this query's scope ends.  This
    * implementation cleans up any open results.
    * 
    * @see  P2JQuery#cleanup()
    */
   public void cleanup()
   {
      if (!closed)
      {
         if (results != null)
         {
            results.cleanup();
         }
         
         closed = true;
      }
   }
   
   /**
    * Open the prepared query and reposition it to the first result if the
    * query is being browsed.
    */
   public void open()
   {
      setErrorIfNull(false);
      setLenientOffEnd(true);
      execute();
      
      super.open();
   }
   
   /**
    * Override parent's no-op implementation to track whether client code has
    * set this as a scrollable query.  While this setting does not explicitly
    * affect this query's ability to respond to reposition requests (this
    * implementation is inherently scrollable), it does provide a hint to the
    * implementation regarding the use and scope of this query.  This method
    * must be invoked before any results are retrieved.
    */
   public void setScrolling()
   {
      scrolling = true;
   }
   
   /**
    * Disable results list scrolling/repositioning (if it was enabled by
    * default). This method must be invoked before any results are retrieved. 
    */
   public void setNonScrolling()
   {
      scrolling = false;
   }
   
   /**
    * Get the off-end status of this query, which indicates whether the query
    * has run off either end of its results.
    * 
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      return offEnd;
   }
   
   /**
    * Indicate whether this query's <code>scrolling</code> flag has been set.
    * 
    * @return  <code>true</code> if this is a scrolling query;  else
    *          <code>false</code>.
    */
   protected boolean isScrolling()
   {
      return scrolling || isNativelyScrolling();
   }
   
   /**
    * Indicate whether this query type is scrolling by its nature.
    * <p>
    * Subclasses which are not natively scrolling should override this method.
    * 
    * @return  <code>true</code>.
    */
   protected boolean isNativelyScrolling()
   {
      return true;
   }

   /**
    * Indicate whether this query is a preselect query by its nature.
    *
    * @return  <code>true</code>.
    */
   protected boolean isNativelyPreselect()
   {
      return true;
   }

   /**
    * Discard the current result set, clear the reposition cache, and reset
    * result-related state to default values.
    */
   protected void resetResults()
   {
      if (results != null)
      {
         results.cleanup();
      }
      setResults(null);
      offEnd = OffEnd.NONE;
      betweenRows = true;
      rowCount = -1;
      clearRepoCache();
   }

   /**
    * Get an iterator on the query components.
    *
    * @return  Iterator on components.
    */
   protected Iterator<Component> components()
   {
      return components.iterator();
   }
   
   /**
    * Indicate whether this query will throw a
    * <code>QueryOffEndException</code> when it runs off the end of its
    * results.
    *
    * @return  <code>true</code> if an exception will be thrown, else
    *          <code>false</code>.
    */
   protected boolean isLenientOffEnd()
   {
      return lenientOffEnd;
   }
   
   /**
    * Indicate whether a call to move to the previous record in the result
    * list should be treated as a call to move to the last record in the list.
    * This will be the case if either there are no results available, or if
    * the query is not off the front end of its results, but the row number
    * is -1 (indicating no results have yet been retrieved).
    * 
    * @return  <code>true</code> to treat a previous request as a request to
    *          move to the last record in the list;  else <code>false</code>.
    */
   protected boolean treatPreviousAsLast()
   {
      return (results == null ||
              (offEnd != OffEnd.FRONT && results.getRowNumber() == -1));
   }
   
   /**
    * Walk through each component and set its backing buffer to unknown mode,
    * such that the invocation of any of its DMO-backed methods returns
    * unknown value.
    */
   protected void setEmptyBuffersUnknown()
   {
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next().getBuffer();
         if (buffer.getCurrentRecord() == null)
         {
            buffer.setUnknownMode();
         }
      }
   }
   
   /**
    * Given a primary DMO buffer and an inverse DMO buffer which together
    * define a natural join, create the <code>AbstractJoin</code> helper
    * object which is used to implement the join.
    * <p>
    * It is intended that this method be called when adding a query component.
    * If no previous components have been added, a <code>DynamicJoin</code>
    * is created, and it is assumed that the inverse DMO is provided as a
    * query substitution parameter.  If a component previously has been added,
    * a <code>ServerJoin</code> is created, and it is assumed that the inverse
    * DMO will be resolved by a server-side join, rather than being provided
    * as a query substitution parameter.
    * <p>
    * If no inverse buffer is provided, no join is created.
    *
    * @param   primary
    *          Buffer reference to the primary DMO.
    * @param   inverse
    *          Buffer reference to the joined DMO, or <code>null</code> if no
    *          join is required.
    *
    * @return  A join helper object, or <code>null</code> if no natural join
    *          is required.
    */
   protected AbstractJoin processJoin(BufferReference primary,
                                      BufferReference inverse)
   {
      if (inverse == null)
      {
         return null;
      }
      
      AbstractJoin join = null;
      
      if (getTableCount() == 0)
      {
         // If we have an inverse proxy and this is the first component being
         // added, the inverse is being provided as a query substitution
         // parameter.
         join = Persistence.isForeignKeysEnabled() ?
                new DynamicJoin(primary.buffer(), inverse.buffer()) :
                new DynamicLegacyKeyJoin(primary.buffer(), inverse.buffer());
      }
      else
      {
         // If we have an inverse proxy and this is not the first component
         // being added, the inverse is provided as a related DMO from a
         // related table.  In this case, the inverse was not provided as a
         // query substitution parameter.
         join = Persistence.isForeignKeysEnabled() ?
                new ServerJoin(primary.buffer(), inverse.buffer()) : 
                new ServerLegacyKeyJoin(primary.buffer(), inverse.buffer());
      }
      
      return join;
   }
   
   /**
    * Assemble the select clause for an HQL projection query, where the only
    * field retrieved for each buffer is the primary key. This uses the
    * Hibernate reserved keyword "id" to represent the primary key. Form of
    * the clause is:
    * <pre>
    *    select {buffer0.id} [, {buffer1.id} [...]]
    * </pre>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    */
   protected void assembleSelectClause(StringBuilder buf)
   {
      boolean needSelect = !fullRecords;
      
      Iterator<Component> iter = components();
      while (!needSelect && iter.hasNext())
      {
         Component comp = iter.next();
         if (!isClientWhere() && comp.isIDOnly())
         {
            needSelect = true;
         }
      }
      
      if (!needSelect)
      {
         return;
      }
      
      buf.append("select ");
      
      iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         Component comp = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         
         RecordBuffer buffer = comp.getBuffer();
         buf.append(buffer.getDMOAlias());
         if (!isClientWhere() && comp.isIDOnly())
         {
            buf.append(".id");
         }
      }
      
      buf.append(" ");
   }
   
   /**
    * Assemble the from clause for an HQL query. The class name of each
    * DMO is decapitalized to produce an alias for that object. Form of
    * the clause is:
    * <pre>
    *    from {Buffer0} as {buffer0} [, {Buffer1} as {buffer1} [, ...]]
    * </pre>
    * <p>
    * <b>Known Limitation:</b>
    * At this time, only theta-style joins are supported and as such, only
    * inner joins are supported (there is no way to specify an outer join
    * using theta-join syntax in HQL).
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which the composite join data are gathered.
    */
   protected void assembleFromClause(StringBuilder buf,
                                     List<SortCriterion> sortCriteria)
   throws PersistenceException
   {
      buf.append("from ");
      
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         Component comp = iter.next();
         RecordBuffer buffer = comp.getBuffer();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(buffer.getDMOImplementationName());
         buf.append(" as ");
         buf.append(buffer.getDMOAlias());
      }
      
      buf.append(" ");
      
      // Add any ANSI joins required by the HQL preprocessor of each
      // component.
      iter = components();
      while (iter.hasNext())
      {
         Component comp = iter.next();
         Iterator<String> joinsIter = comp.compositeJoins();
         while (joinsIter.hasNext())
         {
            buf.append(joinsIter.next());
            buf.append(" ");
         }
      }
      
      // Add any ANSI joins required by sort criteria.
      if (sortCriteria != null)
      {
         Iterator<SortCriterion> sortIter = sortCriteria.iterator();
         while (sortIter.hasNext())
         {
            SortCriterion crit = sortIter.next();
            String join = crit.getAnsiJoin();
            if (join != null)
            {
               buf.append(join);
               buf.append(" ");
            }
         }
      }
   }
   
   /**
    * Assemble the restriction (i.e., where) clause for an HQL query. If more
    * than one component contains a where clause, they are combined with a
    * logical "and". Form of the clause is:
    * <pre>
    *    where ({component0_clause})
    *     [and ({component1_clause})
    *     [...]]
    * </pre>
    * <p>
    * Please refer to the Hibernate reference documentation for permissible
    * syntax of this clause.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which the preprocessed where clause is retrieved for each
    *          query component.
    */
   protected void assembleWhereClause(StringBuilder buf,
                                      List<SortCriterion> sortCriteria)
   throws PersistenceException
   {
      int start = buf.length();
      
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component next = iter.next();
         RecordBuffer rb = next.getBuffer();
         boolean multiplex = rb.isMultiplexed();
         AbstractJoin join = next.getJoin();
         String where = StringHelper.safeTrim(next.getWhere());
         boolean hasJoin = (join != null);
         boolean hasWhere = (where != null && where.length() > 0);
         boolean hasContent = (hasJoin || multiplex || hasWhere);
         
         if (hasContent)
         {
            if (buf.length() == start)
            {
               buf.append("where (");
            }
            else
            {
               buf.append("and (");
            }
         }
         
         // Determine whether where clause should be augmented with an
         // equality test for a multiplex ID.  We need to do this when:
         // a) we are querying a multiplexed temp table;  OR
         // b) we are querying any temp table and have a non-empty where
         //    clause
         if (multiplex)
         {
            if (hasJoin || hasWhere)
            {
               buf.append("(");
            }
            
            buf.append(rb.getDMOAlias());
            buf.append(".");
            buf.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
            buf.append(" = ?");
            
            if (hasJoin || hasWhere)
            {
               buf.append(") and ");
            }
         }
         
         // Add subclause for a join.
         if (hasJoin)
         {
            if (multiplex || hasWhere)
            {
               buf.append("(");
            }
            
            buf.append(join.getHQL());
            
            if (multiplex || hasWhere)
            {
               buf.append(")");
            }
            
            if (hasWhere)
            {
               buf.append(" and ");
            }
         }
         
         if (hasWhere)
         {
            if (hasJoin || multiplex)
            {
               buf.append("(");
            }
            
            buf.append(where);
            
            if (hasJoin || multiplex)
            {
               buf.append(")");
            }
         }
         
         if (hasContent)
         {
            buf.append(") ");
         }
      }
      
      if (sortCriteria != null)
      {
         Iterator<SortCriterion> sortIter = sortCriteria.iterator();
         while (sortIter.hasNext())
         {
            SortCriterion crit = sortIter.next();
            String rest = crit.getCompositeRestrictor();
            if (rest != null)
            {
               if (buf.length() == start)
               {
                  buf.append("where (");
               }
               else
               {
                  buf.append("and (");
               }
               buf.append(rest);
               buf.append(") ");
            }
         }
      }
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query.  Uses the
    * form:
    * <pre>
    *   order by {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if the raw sort clause provided at construction contains any
    *          unqualified property names;  if the raw sort clause contains
    *          a reference to an unrecognized record buffer.
    */
   protected List<SortCriterion> assembleOrderByClause(StringBuilder buf)
   throws PersistenceException
   {
      if (sort == null)
      {
         return null;
      }
      
      buf.append("order by ");
      
      return assembleOrderByClauseNoPreamble(buf, sort);
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query, sans the
    * &quot;order by&quot; preamble.  Uses the
    * form:
    * <pre>
    *   {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sort
    *          Raw sort clause to parse and augment if necessary.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if <code>sort</code> contains any unqualified property names;
    *          if <code>sort</code> clause contains a reference to an
    *          unrecognized record buffer.
    */
   protected List<SortCriterion>
   assembleOrderByClauseNoPreamble(StringBuilder buf, String sort)
   throws PersistenceException
   {
      List<SortCriterion> sortCriteria = buildSortCriteria(sort);

      // Assemble the order by clause, given the list of SortCriterion
      // objects.
      return assembleOrderByClauseNoPreamble(buf, sortCriteria);
   }
   
   /**
    * Returns a list <code>SortCriterion</code> objects build from the given
    * sort clause.
    *
    * @param   sort
    *          Raw sort clause to parse and augment if necessary.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if <code>sort</code> contains any unqualified property names;
    *          if <code>sort</code> clause contains a reference to an
    *          unrecognized record buffer.
    */
   protected List<SortCriterion> buildSortCriteria(String sort)
   throws PersistenceException
   {
      // Map record buffers used in the sort clause to their identifying
      // aliases.
      Map<String, RecordBuffer> aliases = new HashMap<String, RecordBuffer>();
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next().getBuffer();
         aliases.put(buffer.getDMOAlias(), buffer);
      }
      
      // Parse the sort clause, extracting each qualified DMO property
      // reference, matching it to its buffer (by alias), and creating a
      // SortCriterion object for it.  Also group lists of SortCriterion
      // objects by buffer, so that each query component can store its portion
      // of the sort clause.
      Map<RecordBuffer, List<SortCriterion>> subSortMap =
         new HashMap<RecordBuffer, List<SortCriterion>>();
      List<SortCriterion> sortCriteria = new ArrayList<SortCriterion>();
      String[] parts = sort.split(",");
      int len = parts.length;
      for (int i = 0; i < len; i++)
      {
         String part = parts[i].trim();
         int pos = part.indexOf(".");
         if (pos < 0)
         {
            throw new PersistenceException(
               "Order by clause contains unqualified property name:  " +
               part);
         }
         
         // Sanity.
         String alias = part.substring(0, pos);
         RecordBuffer buffer = aliases.get(alias);
         if (buffer == null)
         {
            throw new PersistenceException(
               "Order by clause references an unrecognized buffer:  " +
               alias + " (" + part + ")");
         }
         
         // Find or create the sub-list in which to store SortCriterion
         // objects for the current buffer.
         List<SortCriterion> subSortCrit = subSortMap.get(buffer);
         if (subSortCrit == null)
         {
            subSortCrit = new ArrayList<SortCriterion>();
            subSortMap.put(buffer, subSortCrit);
         }
         
         // Create the SortCriterion instance and add it to the master list
         // and to the appropriate sub-list.
         
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         
         SortCriterion crit =
            new SortCriterion(buffer, part, schema, dmoIface, dmoClass);
         
         sortCriteria.add(crit);
         subSortCrit.add(crit);
      }
      
      // Match the sort criteria sub-list for each buffer, generate a
      // normalized sort clause for that sub-list, and store it in the
      // appropriate query component.
      Iterator<Map.Entry<RecordBuffer, List<SortCriterion>>> iter2 =
         subSortMap.entrySet().iterator();
      while (iter2.hasNext())
      {
         Map.Entry<RecordBuffer, List<SortCriterion>> next = iter2.next();
         RecordBuffer buffer = next.getKey();
         Component comp = findComponent(buffer);
         assert (comp != null);
         List<SortCriterion> subSortCrit = next.getValue();
         String subSort = SortCriterion.toRawSortPhrase(subSortCrit);
         SortIndex sortIndex = SortIndex.get(subSort, buffer);
         comp.setSortIndex(sortIndex);
      }

      return sortCriteria;
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query, sans the
    * &quot;order by&quot; preamble.  Uses the
    * form:
    * <pre>
    *   {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sortCriteria
    *          A list of <code>SortCriterion</code> objects which describe
    *          the sorting to be applied.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    */
   protected List<SortCriterion>
   assembleOrderByClauseNoPreamble(StringBuilder buf,
                                   List<SortCriterion> sortCriteria)
   {
      RecordBuffer buffer = getRecordBuffers()[0];
      boolean multiplex = (buffer.getMultiplexID() != null);
      List<SortCriterion> workList = null;
      
      if (sortCriteria == null)
      {
         if (multiplex)
         {
            workList = new ArrayList<SortCriterion>(1);
         }
         else
         {
            return null;
         }
      }
      
      if (workList == null)
      {
         workList = new ArrayList<SortCriterion>(sortCriteria.size() + 1);
      }
      
      workList.addAll(sortCriteria);
      
      if (multiplex)
      {
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         
         StringBuilder sb = new StringBuilder();
         sb.append(buffer.getDMOAlias());
         sb.append(".");
         sb.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         sb.append(" ");
         boolean ascending = true;
         int insertAt = sortCriteria.size();
         if (!sortCriteria.isEmpty())
         {
            // Make sure multiplex ID sort direction matches that of the first
            // sort component, otherwise the matching index will not be used
            // by the DB.
            SortCriterion crit = sortCriteria.get(0);
            ascending = crit.isAscending();
            
            int last = sortCriteria.size() - 1;
            crit = sortCriteria.get(last);
            if ("id".equals(crit.getUnqualifiedName()))
            {
               // If the primary key is the final sort criterion, we want to
               // insert the multiplex criterion just before it.
               insertAt--;
            }
         }
         sb.append(ascending ? "asc" : "desc");
         try
         {
            SortCriterion multiCrit = new SortCriterion(buffer,
                                                        sb.toString(),
                                                        schema,
                                                        dmoIface,
                                                        dmoClass);
            workList.add(insertAt, multiCrit);
         }
         catch (PersistenceException exc)
         {
            LOG.error("Error adding multiplex ID to order by clause",
                      exc);
         }
      }
      
      Iterator<SortCriterion> iter = workList.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         SortCriterion crit = iter.next();
         
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(crit.toSortExpression(false));
      }
      
      return sortCriteria;
   }
   
   /**
    * Get this query's client-side where expressions.
    *
    * @return  Unmodifiable where expression list.
    */
   protected List<WhereExpression> getWhereExpressions()
   {
      return Collections.unmodifiableList(whereExprList);
   }
   
   /**
    * Get an iterator on this query's client-side where expressions.
    *
    * @return  Where expression iterator.
    */
   protected Iterator<WhereExpression> whereExpressions()
   {
      return getWhereExpressions().iterator();
   }
   
   /**
    * Compose an HQL query statement to represent the combination of all query
    * components, execute it, and store the scrollable results.  If query has
    * been executed previously and we already have a result set, simply
    * return.
    *
    * @return  <code>true</code> if query was executed for the first time;
    *          <code>false</code> if query was executed previously and we
    *          already have a result set.
    *
    * @throws  ErrorConditionException
    *          if an error occurs executing the query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   protected boolean execute()
   {
      if (results != null)
      {
         return false;
      }
      
      if (components.isEmpty())
      {
         throw new IllegalStateException(
            "Preselect query must contain at least one query component");
      }
      
      if (!verifyJoins())
      {
         return true;
      }
      
      Persistence persistence = getPersistence();
      
      if (!regWithSession)
      {
         // Register to receive session closing notification.
         SessionListener.Scope scope =
            (isScrolling() || isTopLevelCleanup() || !isStandalone()
             ? SessionListener.Scope.ENCLOSING_EXTERNAL
             : SessionListener.Scope.CURRENT);
         persistence.registerSessionListener(this, scope);
         regWithSession = true;
      }
      
      try
      {
         // Assemble HQL statement.
         hql = assembleHQL();
         
         // Prepare substitution parameters.
         List<Object> argList = new ArrayList<Object>();
         List<Type> typeList = new ArrayList<Type>();
         prepareParameters(argList, typeList);
         Object[] args = argList.toArray();
         Type[] types = typeList.toArray(new Type[] {});
         DBUtils.makeTypeArray(args, types);
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("HQL:  " + hql);
            StringBuilder buf = new StringBuilder("PARMS:  {");
            for (int i = 0; i < args.length; i++)
            {
               if (i > 0)
               {
                  buf.append(", ");
               }
               Object next = args[i];
               if (next instanceof BaseDataType)
               {
                  buf.append(((BaseDataType) next).toStringMessage());
               }
               else
               {
                  buf.append(String.valueOf(next));
               }
            }
            buf.append("}");
            LOG.debug(buf.toString());
         }
         
         // Possibly leak uncommitted updates to other sessions.
         Iterator<Component> iter = components();
         while (iter.hasNext())
         {
            Component next = iter.next();
            HQLPreprocessor hqlPreproc = next.getHQLPreprocessor();
            List<String> earlyPublish = hqlPreproc.getEarlyPublishEntities();
            if (!earlyPublish.isEmpty())
            {
               RecordBuffer buffer = next.getBuffer();
               DirtyShareContext dirtyContext = buffer.getDirtyContext();
               List<Persistable> dmos =
                  dirtyContext.updateSnapshots(earlyPublish, persistence);
               
               // Make sure DMOs do not bloat the Hibernate session.
               for (Persistable dmo : dmos)
               {
                  buffer.evictDMOIfUnused(dmo);
               }
            }
         }
         
         // Execute the query.
         Results r = executeQuery(persistence, hql, args, types);
         r = createResults(r);
         setResults(r);
         
         registerCleaner();
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         rowCount = -1;
      }
      
      return true;
   }

   /**
    * Indicate whether results resources should be cleaned up upon exiting
    * the global scope or upon exiting the nearest enclosing scope in which
    * the results were created.
    * <p>
    * This default implementation returns <code>false</code>, indicating
    * results should be cleaned up at the nearest enclosing scope.
    * Subclasses which require different behavior must override this method.
    *
    * @return  <code>false</code>.
    */
   protected boolean isTopLevelCleanup()
   {
      return false;
   }
   
   /**
    * Invoked just before query execution, this method checks whether all
    * components with a non-null JOIN have a valid inverse on which to perform
    * the join.  If not, set the query's results to an empty result set and
    * indicate failure.
    * 
    * @return  <code>true</code> if verification succeeded and query execution
    *          should continue;  else <code>false</code>
    */
   protected boolean verifyJoins()
   {
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component comp = iter.next();
         AbstractJoin join = comp.getJoin();
         
         if (join != null && !join.isServerJoin())
         {
            BufferReference ref = join.getInverse();
            Persistable record = ref.buffer().getCurrentRecord();
            
            if (record == null)
            {
               handleMissingJoinRecord();
               
               return false;
            }
         }
      }
      
      return true;
   }
   
   /**
    * Respond to the condition that a component with a non-null JOIN has a
    * missing inverse on which to perform the join.
    * <p>
    * This implementation creates an empty result set.
    */
   protected void handleMissingJoinRecord()
   {
      setResults(new SimpleResults(null));
   }
   
   /**
    * Attempt to cache all results from the specified target result set into
    * a {@link SimpleResults} instance.  If this is a scrolling query, this
    * will cache all results in the target result set, including those already
    * visited.  Otherwise, only the current result and all remaining,
    * unvisited results are cached.
    * 
    * @param   target
    *          Result set which is to be cached, presumably because it will
    *          be invalid momentarily.  If this is already an instance of
    *          <code>SimpleResults</code>, it is simply returned.
    * 
    * @return  Results from <code>target</code>, if any, cached in an instance
    *          of <code>SimpleResults</code>.  If there were no results
    *          available to cache, an empty result set is returned.  If any of
    *          the record buffers backing this query have been closed,
    *          <code>null</code> is returned.
    */
   protected SimpleResults cacheResults(Results target)
   {
      if (target instanceof SimpleResults)
      {
         // Results are already cached.
         return (SimpleResults) target;
      }
      
      // If any buffer associated with this query is closed, do not
      // attempt to cache results.  Instead, clean up resources;  this
      // query is done.
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component next = iter.next();
         RecordBuffer buffer = next.getBuffer();
         if (buffer.getOpenScopeCount() < 1)
         {
            resetResults();
            
            return null;
         }
      }
      
      // Remember current row.
      int currentRow = (isScrolling() ? target.getRowNumber() : 0);
      
      // Cache all preselected results in a SimpleResults object and set
      // that as the query's new backing results.
      ArrayList<Object[]> rows = new ArrayList<Object[]>(64);
      
      if (isScrolling())
      {
         target.reset();
         while (target.next())
         {
            rows.add(getRow());
         }
      }
      else if (target.get() != null)
      {
         do
         {
            rows.add(getRow());
         } while (target.next());
      }
      else if (target.getRowNumber() == -1)
      {
         currentRow = -1;
         while (target.next())
         {
            rows.add(getRow());
         }
      }
      
      rows.trimToSize();
      
      // Override SimpleResults.cleanup() to do nothing, so we don't
      // immediately discard results when the current scope ends and our
      // Finalizable cleaner (installed in execute()) does its work.
      SimpleResults cached = new SimpleResults(rows)
      {
         public void cleanup() {}
      };
      
      // Reset to current row.
      cached.setRowNumber(currentRow);
      
      return cached;
   }
   
   /**
    * Prepare query substitution parameters for this query.  One argument
    * must be added to each of the <code>args</code> and <code>types</code>
    * lists for each <code>?</code> placeholder in the HQL query string for
    * this query.
    *
    * @param   args
    *          Argument list to be populated.
    * @param   types
    *          Type list to be populated.  This may be populated with
    *          <code>null</code> values, but it must coincide in size and in
    *          the absolute positions of its elements with the
    *          <code>args</code> list.  Any <code>null</code> element will
    *          be filled in later.
    * 
    * @throws  PersistenceException
    *          if there is an error creating an <code>HQLPreprocessor</code>
    *          during the gathering of query substitution parameters.
    */
   protected void prepareParameters(List<Object> args, List<Type> types)
   throws PersistenceException
   {
      Iterator<Component> iter = components.iterator();
      while (iter.hasNext())
      {
         Component next = iter.next();
         
         RecordBuffer rb = next.getBuffer();
         if (rb.isMultiplexed())
         {
            args.add(rb.getMultiplexID());
            types.add(null);
         }
         
         AbstractJoin join = next.getJoin();
         if (join != null)
         {
            List<Serializable> parameters = join.getParameters();
            List<Type> parameterTypes = join.getParameterTypes();

            if (parameterTypes != null)
            {
               for (int i = 0; i < parameterTypes.size(); i++)
               {
                  Type paramType = parameterTypes.get(i);
                  if (paramType != null)
                  {
                     args.add(parameters.get(i));
                     types.add(paramType);
                  }
               }
            }
         }
         
         Object[] cArgs = next.getArgs();
         if (cArgs != null)
         {
            int len = cArgs.length;
            for (int i = 0; i < len; i++)
            {
               Object arg = cArgs[i];
               if (arg instanceof Resolvable)
               {
                  args.add(((Resolvable) arg).resolve());
               }
               else
               {
                  args.add(arg);
               }
               types.add(null);
            }
         }
      }
   }
   
   /**
    * Execute the query to create a first level result set.
    * <p>
    * This default implementation returns an object which is a thin wrapper
    * around <code>ScrollableResults</code>.  Subclasses may override this
    * method if they need a <code>Results</code> implementation other than
    * <code>ScrollingResults</code>.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          within the HQL query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  An instance of <code>ScrollingResults</code> which wrappers
    *          the <code>ScrollableResults</code> object returned by the
    *          execution of the query.
    *
    * @throws  PersistenceException
    *          if there is an error executing the query.
    */
   protected Results executeQuery(Persistence persistence,
                                  String hql,
                                  Object[] args,
                                  Type[] types)
   throws PersistenceException
   {
      boolean pop = false;
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         return (new ScrollingResults(persistence,
                                      executeScroll(persistence,
                                                    hql,
                                                    args,
                                                    types)));
      }
      finally
      {
         if (pop)
         {
            persistence.popImplicitTransaction();
         }
      }
   }
   
   
   /**
    * Execute the query to create a <code>ScrollableResults</code> result set.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          within the HQL query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  The <code>ScrollableResults</code> object returned by the
    *          execution of the query.
    *
    * @throws  PersistenceException
    *          if there is an error executing the query.
    */
   protected ScrollableResults executeScroll(Persistence persistence,
                                             String hql,
                                             Object[] args,
                                             Type[] types)
   throws PersistenceException
   {
      return persistence.scroll(hql, args, types);
   }
   
   /**
    * Optionally wrapper the result set created for the query with a fitering
    * <code>Results</code> implementation (for client-side where processing).
    * It is this object which actually will be used to retrieve the primary
    * key values of records to be loaded from the database.
    * <p>
    * If no filtering is necessary, simply return the results object which
    * was passed in.
    * <p>
    * Subclasses may override this method to insert their own layers into
    * result set processing.
    *
    * @param   results
    *          An object which manipulates the database result set returned
    *          from execution of the query.
    *
    * @return  <code>Results</code> object.
    *
    * @throws  PersistenceException
    *          if there is any error creating results.
    */
   protected Results createResults(Results results)
   throws PersistenceException
   {
      if (isClientWhere())
      {
         return (new FilteredResults(results,
                                     Arrays.asList(getRecordBuffers()),
                                     getWhereExpressions(),
                                     !fullRecords));
      }
      
      return results;
   }
   
   /**
    * Fetch all records for a given result row into their backing buffers.
    * Each result row contains one column per query component.  Each column
    * contains the unique ID of the record to be fetched or the record itself.
    * If the latter, the record is refetched in case it has changed since the
    * record was first retrieved.  If the scrollable cursor backing this query
    * currently is at a position where no retrieval is possible, the backing
    * buffers will have their current records released.
    *
    * @param   available
    *          <code>true</code> if the most recent scrollable cursor move
    *          left the cursor at an available result row;  <code>false</code>
    *          if the cursor cannot read a result at its current position.
    * @param   lockType
    *          Overriding lock type to apply to fetched records.  If
    *          <code>null</code>, default lock types are applied.
    * @param   errorIfNull
    *          <code>true</code> if failed retrievals should raise error.
    * 
    * @throws  MissingRecordException
    *          if the record was expected to be available, but wasn't.  This
    *          typically indicates the record was deleted between the time it
    *          initially was found by the query and the time it was to be
    *          fetched to store in the record buffer.
    * @throws  ErrorConditionException
    *          if the fetch of any record fails or if no records are available
    *          and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    */
   protected void fetch(boolean available,
                        LockType lockType,
                        boolean errorIfNull)
   throws MissingRecordException
   {
      try
      {
         Object[] data = (available ? results.get() : null);
         coreFetch(data, lockType, errorIfNull);
         
         if (available)
         {
            accumulate();
         }
      }
      catch (QueryOffEndException exc)
      {
         if (!lenientOffEnd)
         {
            throw exc;
         }
      }
      catch (MissingRecordException exc)
      {
         throw exc;
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Fetch all records for a given result row into their backing buffers,
    * given an array of data which contains one primary key or DMO per query
    * component.
    *
    * @param   data
    *          An array of primary key IDs or DMOs representing records to be
    *          fetched for the composite row.
    * @param   lockType
    *          Overriding lock type to apply to fetched records.  If
    *          <code>null</code>, default lock types are applied.
    * @param   errorIfNull
    *          <code>true</code> if failed retrievals should raise error.
    * 
    * @throws  MissingRecordException
    *          if the record was expected to be available, but wasn't.  This
    *          typically indicates the record was deleted between the time it
    *          initially was found by the query and the time it was to be
    *          fetched to store in the record buffer.
    * @throws  PersistenceException
    *          if the fetch of any record fails.
    * @throws  ErrorConditionException
    *          if no records are available and the query is set to fail in
    *          this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    */
   protected final void coreFetch(Object[] data, LockType lockType, boolean errorIfNull)
   throws MissingRecordException,
          PersistenceException
   {
      boolean debug = LOG.isDebugEnabled();
      
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         Component comp = iter.next();
         RecordBuffer buffer = comp.getBuffer();
         boolean dirtyCopy = false;
         Persistable dmo = null;
         LockType lt = lockType;
         if (data != null)
         {
            LockType compLock = comp.getLockType();
            
            // Use default lock type if not overridden.
            if (lt == null)
            {
               lt = compLock;
            }
            
            // Individual result is either a DMO (i.e., Persistable) or
            // a primary key ID.
            Object o = data[i];
            Serializable id = null;
            boolean updateLock;
            if (o instanceof Persistable)
            {
               dmo = (Persistable) o;
               id = dmo.getId();
               updateLock = !LockType.NONE.equals(lt);
            }
            else
            {
               id = (Serializable) o;
               updateLock = true;
            }
            
            // If the DMO currently in the buffer has the same ID as the ID
            // we wish to load, but also represents a dirty copy (a copy of
            // uncommitted changes in another session), leave it in the buffer.
            if (buffer.isDirtyCopy())
            {
               Persistable dirtyDMO = buffer.getCurrentRecord();
               if (dirtyDMO != null && id.equals(dirtyDMO.getId()))
               {
                  dirtyCopy = true;
                  dmo = dirtyDMO;
               }
            }
            
            if (!dirtyCopy)
            {
               // A record may have been retrieved from a cache of detached
               // objects, OR its state may have been changed (updated or
               // or deleted) since we first retrieved it, so reload it by its ID.
               dmo = getPersistence().load(buffer, id, lt, updateLock, false);
            }
            
            if (debug)
            {
               StringBuilder buf = new StringBuilder("DMO:  ");
               if (id != null)
               {
                  buf.append(id);
               }
               else
               {
                  buf.append("<not found>");
               }
               
               LOG.debug(buf.toString());
               
               if (LOG.isTraceEnabled())
               {
                  buf.setLength(0);
                  buf.append("DMO Detail:  ");
                  buf.append(dmo);
                  
                  LOG.trace(buf.toString());
               }
            }
            
            // If we expected this record to be available, but it wasn't, it
            // must have been deleted.  Likewise, if we know it was deleted in
            // another context, this is an exceptional condition.  Throw an
            // exception so the caller can figure out the appropriate
            // recovery.
            DirtyShareContext dirtyContext = buffer.getDirtyContext();
            String entity = buffer.getEntityName();
            if (dmo == null || dirtyContext.isDirtyDelete(entity, id))
            {
               String table = buffer.getTable();
               RecordIdentifier ident = new RecordIdentifier(table, id);
               
               throw new MissingRecordException(ident);
            }
         }
         
         // TODO:  confirm or fix behavior:  first null record will raise
         // error or end condition;  remaining buffers will not have their
         // current records released.
         SortIndex sortIndex = comp.getSortIndex();
         buffer.setRecord(dmo,
                          lt,
                          errorIfNull,
                          sortIndex,
                          offEnd,
                          inverseSorting,
                          dirtyCopy);  // TODO:  need better dirty check
         if (dmo == null && errorIfNull)
         {
            buffer.errorNotOnFile();
         }
      }
   }
   
   /**
    * Indicate whether this query requires initial results to be filtered
    * through a client-side where clause expression.
    *
    * @return  <code>true</code> if client-side filter is required, else
    *          <code>false</code>.
    */
   protected final boolean isClientWhere()
   {
      return (!getWhereExpressions().isEmpty());
   }
   
   /**
    * Access the current <code>Results</code> object for this query.
    *
    * @return  <code>Results</code> object.
    */
   protected Results getResults()
   {
      execute();
      
      return results;
   }
   
   /**
    * Add a component to the query.
    *
    * @param   comp
    *          Query component to be added.
    *
    * @throws  UnsupportedOperationException
    *          if attempting an outer join (temporary);
    *          if iteration type is not <code>NEXT</code> (temporary);
    *          if the component's DMO's associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   protected void addComponent(Component comp)
   {
      // Clear temporary silent error mode.
      init();
      
      if (isClientWhere())
      {
         throw new IllegalStateException(
            "Cannot have a multi-table query with a client-side where clause");
      }
      
      // TODO:  add support for outer joins.
      if (comp.isOuter())
      {
         throw new UnsupportedOperationException(
            "Outer join not supported with preselect query");
      }
      
      // TODO:  add support for FIRST/LAST iteration types.
      if (comp.getIteration() != NEXT && !supportsNonNextIteration())
      {
         throw new UnsupportedOperationException(
            "Preselect query only supports the NEXT iteration type");
      }
      
      boolean isFirst = components.isEmpty();
      
      RecordBuffer buffer = comp.getBuffer();
      
      if (buffer.getOpenScopeCount() > 0)
      {
         buffer.initialize();
         
         // Flush the record buffer, in case a transient, newly created record
         // is currently in the buffer.
         try
         {
            buffer.flushAll();
         }
         catch (ValidationException exc)
         {
            ErrorManager.throwError(exc);
         }
         
         if (isFirst)
         {
            // Register for retry support.
            TransactionManager.registerFinalizable(this, false);
         }
         // Verify that no cross-connection join is being attempted.
         else if (!getPersistence().equals(buffer.getPersistence()))
         {
            throw new UnsupportedOperationException(
               "Preselect query cannot join across database connections");
         }
      }
      
      // Add the component.
      components.add(comp);
      
      // Update cumulative substitution argument count.
      Object[] args = comp.getRawArgs();
      if (args != null)
      {
         int len = args.length;
         for (int i = 0; !resolvableArgs && i < len; i++)
         {
            if (args[i] instanceof Resolvable)
            {
               resolvableArgs = true;
            }
         }
      }
   }
   
   /**
    * Does this query support non-NEXT iteration components?
    * 
    * @return  <code>false</code>
    */
   protected boolean supportsNonNextIteration()
   {
      return false;
   }
   
   /**
    * Clear the reposition cache of visited, virtual composite rows.
    */
   protected void clearRepoCache()
   {
      repoCache.clear();
   }
   
   /**
    * Get the HQL query which was executed to produce this query's results.
    *
    * @return  HQL query or <code>null</code> if the query has not yet been
    *          executed.
    */
   protected String getHQL()
   {
      return hql;
   }
   
   /**
    * Get the persistence service object used by this query.
    * 
    * @return  Persistence object.
    */
   protected Persistence getPersistence()
   {
      if (persistence == null)
      {
         // Since no cross-connection joins are permitted, it is safe to set
         // the query's Persistence instance using the first component only.
         RecordBuffer buffer = components.get(0).getBuffer();
         persistence = buffer.getPersistence();
      }
      
      return persistence;
   }
   
   /**
    * Register a <code>Finalizable</code> with the transaction manager for
    * resource cleanup when this query's scope ends.
    */
   protected void registerCleaner()
   {
      // If we are not a standalone query, we assume our container will
      // register for cleanup.
      if (regCleaner || !isStandalone())
      {
         return;
      }
      
      Finalizable cleaner = new Finalizable()
      {
         public void finished() { cleanup(); }
         public void iterate() {}
         public void retry() {}
      };
      if (isTopLevelCleanup())
      {
         TransactionManager.registerTopLevelFinalizable(cleaner, true);
      }
      else
      {
         TransactionManager.registerFinalizable(cleaner, false);
      }
      
      regCleaner = true;
   }
   
   /**
    * Find the query component associated with the given record buffer.
    * 
    * @param   buffer
    *          Record buffer.
    * 
    * @return  Associated query component, or <code>null</code> if not found.
    */
   private Component findComponent(RecordBuffer buffer)
   {
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component next = iter.next();
         if (next.getBuffer() == buffer)
         {
            return next;
         }
      }
      
      return null;
   }
   
   /**
    * Resolve all client-side where expression subsitition parameters.  Does
    * nothing if no client-side where expressions have been registered with
    * this query.
    */
   private void resolveClientWhereArgs()
   {
      Iterator<WhereExpression> iter = whereExpressions();
      while (iter.hasNext())
      {
         WhereExpression next = iter.next();
         next.resolveArgs();
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   ids
    *          Array of primary key IDs, arranged from left to right to
    *          coincide with the records being joined by the underlying
    *          query.
    * @param   forceFetch
    *          <code>true</code> to override the current fetch-on-reposition
    *          setting to <code>true</code>, in case the underlying
    *          <code>Results</code> implementation attempts to fetch the
    *          record during repositioning;  <code>false</code> to honor the
    *          current setting.
    * 
    * @return  <code>true</code> if reposition was successful;
    *          <code>false</code> if unsuccessful.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   private boolean repositionByID(Serializable[] ids, boolean forceFetch)
   {
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }
      
      int newRow = repoCache.getRow(ids);
      if (newRow >= 0)
      {
         int oldRow = results.getRowNumber();
         boolean avail;
         if (forceFetch)
         {
            boolean oldFetch = isFetchOnReposition();
            try
            {
               setFetchOnReposition(true);
               avail = results.setRowNumber(newRow);
            }
            finally
            {
               setFetchOnReposition(oldFetch);
            }
         }
         else
         {
            avail = results.setRowNumber(newRow);
         }
         if (avail)
         {
            offEnd = OffEnd.NONE;
            betweenRows = true;
         }
         else
         {
            int diff = oldRow - newRow;
            offEnd = (diff < 0 ? OffEnd.FRONT : OffEnd.BACK);
            betweenRows = false;
         }
         
         afterReposition(true);
         
         return true;
      }
      
      // Scan all results for a match.  Start from where cache last left off.
      boolean error = false;
      int size = repoCache.size();
      results.setRowNumber(size);
      if (results.get(0) == null)
      {
         repoCache.clear();
         error = !results.first();
      }
      
      int width = getTableCount();
      if (!error)
      {
         do
         {
            boolean match = true;
            Serializable[] rsIDs = new Serializable[width];
            for (int i = 0; i < width; i++)
            {
               Serializable nextID = results.getID(i);
               rsIDs[i] = nextID;
               if (ids.length > i && !ids[i].equals(nextID))
               {
                  match = false;
               }
            }
            
            newRow = results.getRowNumber();
            repoCache.add(rsIDs, newRow);
            
            if (match)
            {
               offEnd = OffEnd.NONE;
               betweenRows = true;
               afterReposition(true);
               
               return true;
            }
            
         } while (results.next());
      }
      
      // No match was found;  reset to old current row.
//      results.setRowNumber(currentRow);
      
      // Here we need to emulate a quirk/defect of Progress' reposition
      // implementation:  if no match was found, we leave the "current"
      // position at the last checked position.  Furthermore, if we are
      // fetching on reposition, we have to load this last record into the
      // associated buffer(s), to truly emulate the GET NEXT semantic.
      if (isFetchOnReposition() && results.last())
      {
         try
         {
            offEnd = OffEnd.NONE;
            fetch(true, null, true);
            afterReposition(false);    // notify listeners
         }
         catch (MissingRecordException exc)
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Error fetching record(s) during reposition", exc);
            }
         }
         catch (ErrorConditionException exc)
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Error fetching record(s) during reposition", exc);
            }
         }
      }
      
      return false;
   }
   
   /**
    * Assemble the HQL which is submitted to Hibernate in order to execute
    * the database query underlying this object. The HQL generated uses the
    * theta join style and includes a select clause, a from clause, an
    * optional where clause, and an order by clause (mandatory in all cases
    * except a unique find).
    *
    * @return  String
    *          A formatted HQL statement.
    *
    * @throws  PersistenceException
    *          if the sort criteria reference unqualified or unrecognized
    *          buffer names.
    */
   private String assembleHQL()
   throws PersistenceException
   {
      StringBuilder buf = new StringBuilder();
      
      // Need to put together order by clause first, as the sort criteria
      // determined in this analysis may be needed when composing from clause
      // and where clause.
      List<SortCriterion> sortCriteria = assembleOrderByClause(buf);
      String orderBy = buf.toString();
      buf.setLength(0);
      
      assembleSelectClause(buf);
      assembleFromClause(buf, sortCriteria);
      assembleWhereClause(buf, sortCriteria);
      buf.append(orderBy);
      
      return buf.toString().trim();
   }
   
   /**
    * Validate a row identifier provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionByID(rowid value)
   {
      String msg = "Could not evaluate rowid during REPOSITION";
      
      return validateRepositionImpl(value, 7330, msg);
   }
   
   /**
    * Validate a record identifier provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionByID(recid value)
   {
      String msg = "Could not evaluate rowid during REPOSITION";
      
      return validateRepositionImpl(value, 7330, msg);
   }
   
   /**
    * Validate a numeric value provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID or row number/offset to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateReposition(NumberType value)
   {
      String msg = "Could not evaluate reposition amount for query";
      
      return validateRepositionImpl(value, 3165, msg);
   }
   
   /**
    * Validate a numeric value provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID or row number/offset to be validated.
    * @param   errorNum
    *          Error number if validation fails.
    * @param   errorMsg
    *          Error message if validation fails.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionImpl(BaseDataType value,
                                          int errorNum,
                                          String errorMsg)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(errorNum, errorMsg, false);
         
         if (LOG.isErrorEnabled())
         {
            LOG.error(errorMsg);
         }
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Prepare to process a reposition request by releasing the current
    * records, if any, in each buffer managed by the query.  This occurs in
    * Progress unconditionally, even if the reposition does not actually move
    * the cursor from its current position.
    */
   private void prepareReposition()
   {
      // Don't release buffer if we're processing a recursive reposition that
      // was triggered by a reposition notification.
      if (isRepositionNotificationActive())
      {
         return;
      }
      
      Iterator<Component> iter = components.iterator();
      while (iter.hasNext())
      {
         Component next = iter.next();
         next.getBuffer().release();
      }
   }
   
   /**
    * A helper container for a query component. A query component defines
    * the buffer to be updated, the type of join to perform with the previous
    * component (if any), the lock type for records fetched into the buffer,
    * the where clause to be applied, and any query substitution parameters.
    */
   protected class Component
   {
      /** Original where clause, <i>before</i> preprocessing */
      private final String where;
      
      /** Record buffer updated by this component of the query */
      private final RecordBuffer buffer;
      
      /** Helper object for a join via a foreign key relation */
      private final AbstractJoin join;
      
      /** Record lock type */
      private final LockType lockType;
      
      /** Base substitution arguments for this component */
      private final Object[] args;
      
      /** Iteration/retrieval mode (first, last, next, etc.) */
      private int iteration;
      
      /**
       * <code>true</code> if this component forms an outer join with the
       * previous component;  <code>false</code> if an inner join, or if there
       * is no previous component.
       */
      private final boolean outer;
      
      /** Whether query should return primary key ID only, or entire DMO */
      private final boolean idOnly;
      
      /** HQL preprocessor */
      private HQLPreprocessor hqlPreprocessor;
      
      /** Sort index associated with this query component */
      private SortIndex sortIndex;
      
      /** Intermediately resolved arguments for this component */
      private Object[] currentArgs = null;
      
      /**
       * Constructor.
       *
       * @param   proxy
       *          DMO proxy which references the record buffer updated by
       *          this component of the query.
       * @param   join
       *          Helper object for a join via a foreign key relation.  May
       *          be <code>null</code>.
       * @param   where
       *          Restriction criteria in HQL where clause format.
       * @param   lockType
       *          Requested record lock type.
       * @param   args
       *          Default (unresolved) query substitution arguments.
       * @param   iteration
       *          Type of iteration:  FIRST, LAST, NEXT, PREVIOUS, or UNIQUE.
       * @param   outer
       *          <code>true</code> if this component is joined with the
       *          previous component via an outer join;  <code>false</code>
       *          if there is no previous component, or if this component is
       *          joined with the previous component via an inner join.
       */
      Component(BufferReference proxy,
                AbstractJoin join,
                String where,
                LockType lockType,
                Object[] args,
                int iteration,
                boolean outer)
      {
         if (lockType == null)
         {
            throw new NullPointerException("Lock type must be specified");
         }
         
         validateSubstitutionArguments(args);
         
         this.buffer = proxy.buffer();
         this.join = join;
         this.where = where;
         this.args = args;
         this.iteration = iteration;
         this.outer = outer;
         
         // Override lock type if this is a temporary buffer;  this allows
         // some downstream optimizations.
         this.lockType = (buffer.isTemporary() ? LockType.NONE : lockType);
         
         this.idOnly = true;
      }
      
      /**
       * Get the HQL preprocessor associated with this component.
       * 
       * @return  HQL preprocessor.
       * 
       * @throws  PersistenceException
       *          if the preprocessor must first be created, and there is an
       *          error during its creation.
       */
      protected HQLPreprocessor getHQLPreprocessor()
      throws PersistenceException
      {
         if (hqlPreprocessor == null)
         {
            Object[] resolvedArgs = resolveArgs();
            hqlPreprocessor = prepareHQLPreprocessor(resolvedArgs);
         }
         
         return hqlPreprocessor;
      }
      
      /**
       * Create the <code>HQLPreprocessor</code> which backs this component.
       * 
       * @param   queryArgs
       *          Query substitution parameters used in the preprocessor's
       *          creation.
       * 
       * @return  HQL preprocessor.
       * 
       * @throws  PersistenceException
       *          if there is an error creating the preprocessor.
       */
      protected HQLPreprocessor
      prepareHQLPreprocessor(Object[] queryArgs)
      throws PersistenceException
      {
         return HQLPreprocessor.get(where, buffer, queryArgs, true);
      }
      
      /**
       * Get substitution arguments for this component, resolving them first
       * if necessary.
       *
       * @return  Substitution arguments.
       * 
       * @throws  PersistenceException
       *          if there is any error creating the HQL preprocessor, which
       *          is necessary to properly filter/order the arguments.
       */
      protected Object[] getArgs()
      throws PersistenceException
      {
         if (args != null && currentArgs == null)
         {
            currentArgs = filterArgs(getHQLPreprocessor());
         }
         
         return currentArgs;
      }
      
      /**
       * Reset current arguments, so that they will need to be resolved the
       * next time they are requested.
       */
      protected void resetArgs()
      {
         currentArgs = null;
         
         // Since the arguments were reset, the HQL preprocessor may need to
         // be reset, too.
         if (hqlPreprocessor != null && hqlPreprocessor.wasInlined())
         {
            resetHQLPreprocessor();
         }
      }
      
      /**
       * Reset the HQL preprocessor, so that it needs to be recreated the next
       * time it is requested.
       */
      protected void resetHQLPreprocessor()
      {
         hqlPreprocessor = null;
      }
      
      /**
       * Get the (mostly) resolved query substitution parameters required by
       * this query component and return them in an array.  The array's
       * elements will match the positional placeholders in the query's where
       * clause <em>after</em> it has been preprocessed.  The preprocessing
       * step may have reordered or dropped certain parameters.  The only
       * parameters which will not have been fully resolved will be those
       * <code>FieldReference</code> parameters which are actually resolved by
       * the query itself, when the query's full HQL is being assembled and
       * when the final array of substitution parameters is prepared.
       * 
       * @param   preproc
       *          HQL preprocessor which maps parameter indices to the
       *          original parameters, in case any were changed during HQL
       *          where clause preprocessing.
       * 
       * @return  Array of filtered, resolved parameters.
       */
      protected Object[] filterArgs(HQLPreprocessor preproc)
      {
         Object[] intermediate = resolveArgs();
         if (intermediate == null)
         {
            return null;
         }
         
         ParameterIndices pi = preproc.getParameterIndices();
         int len = pi.getCount();
         Object[] filteredArgs = new Object[len];
         for (int i = 0, count = 0; i < intermediate.length; i++)
         {
            Integer nextIndex = pi.getIndex(i);
            if (nextIndex == null)
            {
               // Original substitution parameter was inlined into HQL by
               // preprocessor;  skip it.
               continue;
            }
            
            // Substitution placeholders may have been reordered if this
            // statement was rewritten by the HQL preprocessor. Make sure
            // we access the correct substitution parameter.
            Object next = intermediate[nextIndex];
            
            filteredArgs[count++] = next;
         }
         
         return filteredArgs;
      }
      
      /**
       * Resolve all resolvable query substitution parameters, with the
       * exception of <code>FieldReference</code>s.  The latter are only
       * query substitution parameters when it is necessary to join the
       * enclosing query's results with those of a related query, as would be
       * the case when each query is a component of a {@link CompoundQuery}.
       * 
       * @return  Array of query substitution parameters after all
       *          non-<code>FieldReference</code> resolvable parameters have
       *          been resolved.
       */
      private Object[] resolveArgs()
      {
         if (args == null)
         {
            return null;
         }
         
         int len = args.length;
         Object[] resolvedArgs = new Object[len];
         
         for (int i = 0; i < len; i++)
         {
            Object next = args[i];
            
            if (next instanceof FieldReference)
            {
               resolvedArgs[i] = next;
            }
            else if (next instanceof Resolvable)
            {
               resolvedArgs[i] = ((Resolvable) next).resolve();
            }
            else
            {
               resolvedArgs[i] = ((BaseDataType) next).duplicate();
            }
         }
         
         return resolvedArgs;
      }
      
      /**
       * Get record buffer for this component of the query.
       *
       * @return  Record buffer.
       */
      RecordBuffer getBuffer()
      {
         return buffer;
      }
      
      /**
       * Get restriction criteria in HQL where clause format for this
       * component (before the HQL preprocessor has preprocessed it).
       *
       * @return  Un-preprocessed where clause.
       */
      String getOriginalWhere()
      {
         return where;
      }
      
      /**
       * Get helper object for a join via a foreign key relation, if any.
       *
       * @return  Join helper, or <code>null</code> if there is none.
       */
      AbstractJoin getJoin()
      {
         return join;
      }
      
      /**
       * Get restriction criteria in HQL where clause format for this
       * component (pre-processed).
       *
       * @return  Where clause.
       * 
       * @throws  PersistenceException
       *          if there is an error creating the
       *          <code>HQLPreprocessor</code> from which the preprocessed
       *          where clause is retrieved.
       */
      String getWhere()
      throws PersistenceException
      {
         return getHQLPreprocessor().getHQL();
      }
      
      /**
       * Get lock type for records fetched into this component's buffer.
       *
       * @return  Lock type.
       */
      LockType getLockType()
      {
         return lockType;
      }
      
      /**
       * Get unresolved substitution arguments for this component, if any.
       *
       * @return  Substitution arguments, or <code>null</code> if none.
       */
      Object[] getRawArgs()
      {
         return args;
      }
      
      /**
       * Get iteration/retrieval mode (first, last, next, etc.).
       *
       * @return  Retrieval mode from {@link QueryConstants}.
       */
      int getIteration()
      {
         return iteration;
      }
      
      /**
       * Set iteration/retrieval mode (first, last, next, etc.).
       *
       * @param   iteration
       *          Retrieval mode from {@link QueryConstants}.
       */
      void setIteration(int iteration)
      {
         this.iteration = iteration;
      }
      
      /**
       * Indicate whether this component is joined with the previous component
       * via a left outer join.
       *
       * @return  <code>true</code> if an outer join, <code>false</code> if
       *          an inner join or no join (i.e., no previous component).
       */
      boolean isOuter()
      {
         return outer;
      }
      
      /**
       * Indicate whether the query component should return a primary key ID
       * only, or the entire DMO instance.
       *
       * @return  <code>true</code> for a projection query which returns only
       *          the ID;  <code>false</code> to fetch the entire DMO.
       */
      boolean isIDOnly()
      {
         return (!fullRecords && idOnly);
      }
      
      /**
       * Get an iterator on the ANSI-style join subexpressions which must be
       * merged by the enclosing query into the overall HQL query statement
       * in order to properly join associated composite element lists (if
       * any).
       *
       * @return  An iterator on join subexpressions.  The iterator may be
       *          empty, but will not be <code>null</code>.
       * 
       * @throws  PersistenceException
       *          if there is an error creating the
       *          <code>HQLPreprocessor</code> from which the composite join
       *          subexpressions are gathered.
       */
      Iterator<String> compositeJoins()
      throws PersistenceException
      {
         return getHQLPreprocessor().ansiJoins();
      }
      
      /**
       * Get an unmodifiable map of DMO entity names to sets of the names of
       * the DMO properties used as restriction criteria in the preprocessed
       * where clause.
       *
       * @return  Map of DMO entity names to sets of property names, or
       *          <code>null</code> if the where clause for this component
       *          contained no restriction properties.
       * 
       * @throws  PersistenceException
       *          if there is an error creating the
       *          <code>HQLPreprocessor</code> from which the restriction
       *          properties are gathered.
       */
      Map<String, Set<String>> getRestrictionProperties()
      throws PersistenceException
      {
         return getHQLPreprocessor().getRestrictionProperties();
      }
      
      /**
       * Get sort criteria in HQL order by clause format for this component.
       *
       * @return  Sort index associated with this component.
       */
      SortIndex getSortIndex()
      {
         return sortIndex;
      }
      
      /**
       * Set sort criteria in HQL order by clause format for this component.
       * 
       * @param   sortIndex
       *          Sort index.
       */
      void setSortIndex(SortIndex sortIndex)
      {
         this.sortIndex = sortIndex;
      }
   }
}
